"""Concrete models.

They mostly inherit from classes in :mod:`.base`.

"""

from collections import OrderedDict, defaultdict
from datetime import datetime
from django.utils import timezone
from decimal import Decimal

from django.conf import settings
from django.urls import reverse
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.template import Context
from django.utils.duration import duration_string

from .base import (
    ArtistBase, IPIBase, LabelBase, LibraryBase, PersonBase, ReleaseBase,
    TitleBase, WriterBase, upload_to
)
from .cwr_templates import (
    TEMPLATES_21, TEMPLATES_22, TEMPLATES_30, TEMPLATES_31)
from .validators import CWRFieldValidator
from .societies import SOCIETIES, SOCIETY_DICT

import base64, uuid

WORLD_DICT = {
    'tis-a': '2WL',
    'tis-n': '2136',
    'name': 'World'
}


class Artist(ArtistBase):
    """Performing artist.
    """

    def get_dict(self):
        """Get the object in an internal dictionary format

        Returns:
            dict: internal dict format
        """
        return {
            'id': self.id,
            'code': self.artist_id,
            'last_name': self.last_name,
            'first_name': self.first_name or None,
            'isni': self.isni or None,
        }

    @property
    def artist_id(self):
        """Artist identifier

        Returns:
            str: Artist ID
        """
        return 'A{:06d}'.format(self.id)


class Label(LabelBase):
    """Music Label.
    """

    class Meta:
        verbose_name = 'Music Label'

    def __str__(self):
        return self.name.upper()

    @property
    def label_id(self):
        """Label identifier

        Returns:
            str: Label ID
        """
        return 'LA{:06d}'.format(self.id)

    def get_dict(self):
        """Get the object in an internal dictionary format

        Returns:
            dict: internal dict format
        """
        return {
            'id': self.id,
            'code': self.label_id,
            'name': self.name,
        }


class Library(LibraryBase):
    """Music Library.
    """

    class Meta:
        verbose_name = 'Music Library'
        verbose_name_plural = 'Music Libraries'
        ordering = ('name',)

    # name = models.CharField(
    #     max_length=60, unique=True,
    #     validators=(CWRFieldValidator('library'),))

    def __str__(self):
        return self.name.upper()

    @property
    def library_id(self):
        """Library identifier

        Returns:
            str: Library ID
        """
        return 'LI{:06d}'.format(self.id)

    def get_dict(self):
        """Get the object in an internal dictionary format

        Returns:
            dict: internal dict format
        """
        return {
            'id': self.id,
            'code': self.library_id,
            'name': self.name,
        }


class Release(ReleaseBase):
    """Music Release (album / other product)

    Attributes:
        library (django.db.models.ForeignKey): Foreign key to \
        :class:`.models.Library`
        release_label (django.db.models.ForeignKey): Foreign key to \
        :class:`.models.Label`
        recordings (django.db.models.ManyToManyField): M2M to \
        :class:`.models.Recording` through :class:`.models.Track`
    """

    class Meta:
        verbose_name = 'Release'

    library = models.ForeignKey(
        Library, null=True, blank=True, on_delete=models.PROTECT)
    release_label = models.ForeignKey(
        Label, verbose_name='Release (album) label', null=True, blank=True,
        on_delete=models.PROTECT)
    recordings = models.ManyToManyField(
        'Recording', through='Track')

    def __str__(self):
        if self.cd_identifier:
            if self.release_title:
                return '{}: {} ({})'.format(
                    self.cd_identifier,
                    self.release_title.upper(),
                    self.library)
            else:
                return '{} ({})'.format(
                    self.cd_identifier, self.library)
        else:
            if self.release_label:
                return '{} ({})'.format(
                    (self.release_title or '<no title>').upper(),
                    self.release_label)
            return (self.release_title or '<no title>').upper()

    @property
    def release_id(self):
        """Release identifier.

        Returns:
            str: Release ID
        """
        return 'RE{:06d}'.format(self.id)

    def get_dict(self, with_tracks=False):
        """Get the object in an internal dictionary format

        Args:
            with_tracks (bool): add track data to the output

        Returns:
            dict: internal dict format

        """

        d = {
            'id': self.id,
            'code': self.release_id,
            'title':
                self.release_title or None,
            'date':
                self.release_date.strftime('%Y%m%d') if self.release_date
                else None,
            'label':
                self.release_label.get_dict() if self.release_label else None,
            'ean':
                self.ean,
        }
        if with_tracks:
            d['tracks'] = [track.get_dict() for track in self.tracks.all()]
        return d


class LibraryReleaseManager(models.Manager):
    """Manager for a proxy class :class:`.models.LibraryRelease`
    """

    def get_queryset(self):
        """Return only library releases

        Returns:
            django.db.models.query.QuerySet: Queryset with instances of \
            :class:`.models.LibraryRelease`
        """
        return super().get_queryset().filter(
            cd_identifier__isnull=False, library__isnull=False)

    def get_dict(self, qs):
        """Get the object in an internal dictionary format

        Args:
            qs (django.db.models.query.QuerySet)

        Returns:
            dict: internal dict format
        """
        return {
            'releases': [release.get_dict(with_tracks=True) for release in qs]
        }


class LibraryRelease(Release):
    """Proxy class for Library Releases (AKA Library CDs)

    Attributes:
        objects (LibraryReleaseManager): Database Manager
    """

    class Meta:
        proxy = True
        verbose_name = 'Library Release'
        verbose_name_plural = 'Library Releases'

    objects = LibraryReleaseManager()

    def clean(self):
        """Make sure that release title is required if one of the other \
        "non-library" fields is present.

        Raises:
            ValidationError: If not compliant.
        """
        if ((self.ean or self.release_date or self.release_label)
                and not self.release_title):
            raise ValidationError({
                'release_title': 'Required if other release data is set.'
            })

    def get_origin_dict(self):
        """Get the object in an internal dictionary format.

        This is used for work origin, not release data.

        Returns:
            dict: internal dict format
        """
        return {
            'origin_type': {
                'code': 'LIB',
                'name': 'Library Work'
            },
            'cd_identifier': self.cd_identifier,
            'library': self.library.get_dict()
        }


class CommercialReleaseManager(models.Manager):
    """Manager for a proxy class :class:`.models.CommercialRelease`
    """

    def get_queryset(self):
        """Return only commercial releases

        Returns:
            django.db.models.query.QuerySet: Queryset with instances of \
            :class:`.models.CommercialRelease`
        """
        return super().get_queryset().filter(
            cd_identifier__isnull=True, library__isnull=True)

    def get_dict(self, qs):
        """Get the object in an internal dictionary format

        Args:
            qs (django.db.models.query.QuerySet)

        Returns:
            dict: internal dict format
        """
        return {
            'releases': [release.get_dict(with_tracks=True) for release in qs]
        }


class CommercialRelease(Release):
    """Proxy class for Commercial Releases

    Attributes:
        objects (CommercialReleaseManager): Database Manager
    """

    class Meta:
        proxy = True
        verbose_name = 'Commercial Release'
        verbose_name_plural = 'Commercial Releases'

    objects = CommercialReleaseManager()


class PlaylistManager(models.Manager):
    """Manager for a proxy class :class:`.models.Playlist`
    """

    def get_queryset(self):
        """Return only commercial releases

        Returns:
            django.db.models.query.QuerySet: Queryset with instances of \
            :class:`.models.CommercialRelease`
        """
        return super().get_queryset().filter(
            cd_identifier__isnull=False, library__isnull=True)

    def get_dict(self, qs):
        """Get the object in an internal dictionary format

        Args:
            qs (django.db.models.query.QuerySet)

        Returns:
            dict: internal dict format
        """
        return {
            'releases': [release.get_dict(with_tracks=True) for release in qs]
        }


class Playlist(Release):
    """Proxy class for Playlists

    Attributes:
        objects (CommercialReleaseManager): Database Manager
    """

    class Meta:
        proxy = True
        verbose_name = 'Playlist'
        verbose_name_plural = 'Playlists'

    objects = PlaylistManager()

    def __str__(self):
        return self.release_title

    def clean(self, *args, **kwargs):
        if self.cd_identifier is None:
            self.cd_identifier = base64.urlsafe_b64encode(uuid.uuid4().bytes)
            self.cd_identifier = self.cd_identifier.decode().rstrip('=')[:15]
        return super().clean(*args, **kwargs)

    @property
    def secret_url(self):
        return reverse('secret_playlist', args=[self.cd_identifier])

    @property
    def secret_api_url(self):
        return reverse('playlist-detail', args=[self.cd_identifier])


class Writer(WriterBase):
    """Writers.

    Attributes:
        original_publishing_agreement (django.db.models.ForeignKey): \
        Foreign key to :class:`.models.OriginalPublishingAgreement`
    """

    class Meta:
        ordering = ('last_name', 'first_name', 'ipi_name', '-id')
        verbose_name = 'Writer'
        verbose_name_plural = 'Writers'

    def __str__(self):
        name = super().__str__()
        if self.generally_controlled:
            return name + ' (*)'
        return name

    def clean(self, *args, **kwargs):
        """Check if writer who is controlled still has enough data."""
        super().clean(*args, **kwargs)
        if self.pk is None or self._can_be_controlled:
            return
        # A controlled writer requires more data, so once a writer is in
        # that role, it is not allowed to remove required data."""
        if self.writerinwork_set.filter(controlled=True).exists():
            raise ValidationError(
                'This writer is controlled in at least one work. ' +
                'Required fields are: Last name, IPI name and PR society. ' +
                'See "Writers" in the user manual.')

    @property
    def writer_id(self):
        """
        Writer ID for CWR

        Returns:
            str: formatted writer ID
        """
        return 'W{:06d}'.format(self.id)

    def get_dict(self):
        """Create a data structure that can be serialized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """

        d = {
            'id': self.id,
            'code': self.writer_id,
            'first_name': self.first_name or None,
            'last_name': self.last_name or None,
            'ipi_name_number': self.ipi_name or None,
            'ipi_base_number': self.ipi_base or None,
            'affiliations': [],
        }
        if self.pr_society:
            d['affiliations'].append({
                'organization': {
                    'code': self.pr_society,
                    'name': self.get_pr_society_display().split(',')[0],
                },
                'affiliation_type': {
                    'code': 'PR',
                    'name': 'Performance Rights'
                },
                'territory': WORLD_DICT,
            })
        if self.mr_society:
            d['affiliations'].append({
                'organization': {
                    'code': self.mr_society,
                    'name': self.get_mr_society_display().split(',')[0],
                },
                'affiliation_type': {
                    'code': 'MR',
                    'name': 'Mechanical Rights'
                },
                'territory': WORLD_DICT,
            })
        if self.sr_society:
            d['affiliations'].append({
                'organization': {
                    'code': self.sr_society,
                    'name': self.get_sr_society_display().split(',')[0],
                },
                'affiliation_type': {
                    'code': 'SR',
                    'name': 'Synchronization Rights'
                },
                'territory': WORLD_DICT,
            })
        return d


class WorkManager(models.Manager):
    """Manager for class :class:`.models.Work`
    """

    def get_queryset(self):
        """
        Get an optimized queryset.

        Returns:
            django.db.models.query.QuerySet: Queryset with instances of \
            :class:`.models.Work`
        """
        return super().get_queryset().prefetch_related('writers')

    def get_dict(self, qs):
        """
        Return a dictionary with works from the queryset

        Args:
            qs(django.db.models.query import QuerySet)

        Returns:
            dict: dictionary with works

        """
        qs = qs.prefetch_related('alternatetitle_set')
        qs = qs.prefetch_related('writerinwork_set__writer')
        qs = qs.prefetch_related('artistinwork_set__artist')
        qs = qs.prefetch_related('library_release__library')
        qs = qs.prefetch_related('recordings__record_label')
        qs = qs.prefetch_related('recordings__artist')
        qs = qs.prefetch_related('recordings__tracks__release__library')
        qs = qs.prefetch_related('recordings__tracks__release__release_label')
        qs = qs.prefetch_related('workacknowledgement_set')

        works = []

        for work in qs:
            j = work.get_dict()
            works.append(j)

        return {
            'works': works,
        }


class Work(TitleBase):
    """Concrete class, with references to foreign objects.

    Attributes:
        _work_id (django.db.models.CharField): permanent work id, either \
        imported or fixed when exports are created
        iswc (django.db.models.CharField): ISWC
        original_title (django.db.models.CharField): title of the original \
            work, implies modified work
        release_label (django.db.models.ForeignKey): Foreign key to \
            :class:`.models.LibraryRelease`
        last_change (django.db.models.DateTimeField):
            when the last change was made to this object or any of the child
            objects, basically used in filtering
        artists (django.db.models.ManyToManyField):
            Artists performing the work
        writers (django.db.models.ManyToManyField):
            Writers who created the work
        objects (WorkManager): Database Manager
    """

    class Meta:
        verbose_name = 'Musical Work'
        ordering = ('-id',)
        permissions = (
            ('can_process_royalties', 'Can perform royalty calculations'),)

    @staticmethod
    def persist_work_ids(qs):
        qs = qs.prefetch_related('recordings')
        for work in qs.filter(_work_id__isnull=True):
            work.work_id = work.work_id
            work.save()
            for rec in work.recordings.all():
                if rec._recording_id is None:
                    rec.recording_id = rec.recording_id

    _work_id = models.CharField(
        'Work ID', max_length=14, blank=True, null=True, unique=True,
        editable=False,
        validators=(CWRFieldValidator('name'),))
    iswc = models.CharField(
        'ISWC', max_length=15, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('iswc'),))
    original_title = models.CharField(
        verbose_name='Title of original work',
        max_length=60, db_index=True, blank=True,
        help_text='Use only for modification of existing works.',
        validators=(CWRFieldValidator('title'),))
    library_release = models.ForeignKey(
        'LibraryRelease', on_delete=models.PROTECT, blank=True, null=True,
        related_name='works',
        verbose_name='Library release')
    last_change = models.DateTimeField(
        'Last Edited', editable=False, null=True)
    artists = models.ManyToManyField('Artist', through='ArtistInWork')
    writers = models.ManyToManyField(
        'Writer', through='WriterInWork', related_name='works')

    objects = WorkManager()

    @property
    def work_id(self):
        """Create Work ID used in registrations.

        Returns:
            str: Internal Work ID
        """
        if self._work_id:
            return self._work_id
        if self.id is None:
            return ''
        return '{}{:06}'.format(settings.PUBLISHER_CODE, self.id)

    @work_id.setter
    def work_id(self, value):
        assert self._work_id is None  # this should not be called if set
        if value:
            self._work_id = value

    def is_modification(self):
        """
        Check if the work is a modification.

        Returns:
            bool: True if modification, False if original
        """
        return bool(self.original_title)

    def clean_fields(self, *args, **kwargs):
        """Deal with various ways ISWC is written.
        """
        if self.iswc:
            # CWR 2.x holds ISWC in TNNNNNNNNNC format
            # CWR 3.0 holds ISWC in T-NNNNNNNNN-C format
            # sometimes it comes in T-NNN.NNN.NNN-C format
            self.iswc = self.iswc.replace('-', '').replace('.', '')
        return super().clean_fields(*args, **kwargs)

    def writer_last_names(self):
        writers = sorted(set(self.writers.all()), key=lambda w: w.last_name)
        return ' / '.join(w.last_name.upper() for w in writers)

    def __str__(self):
        return '{}: {} ({})'.format(
            self.work_id,
            self.title.upper(),
            self.writer_last_names())

    @staticmethod
    def get_publisher_dict():
        """Create data structure for the publisher.

        Returns:
            dict: JSON-serializable data structure
        """
        j = {
            'id': 1,
            'code': settings.PUBLISHER_CODE,
            'name': settings.PUBLISHER_NAME,
            'ipi_name_number': settings.PUBLISHER_IPI_NAME,
            'ipi_base_number': settings.PUBLISHER_IPI_BASE,
            'affiliations': [{
                'organization': {
                    'code': settings.PUBLISHER_SOCIETY_PR,
                    'name': SOCIETY_DICT.get(
                        settings.PUBLISHER_SOCIETY_PR,
                        ''
                    ).split(',')[0],
                },
                'affiliation_type': {
                    'code': 'PR',
                    'name': 'Performance Rights'
                },
                'territory': WORLD_DICT,
            }]
        }

        # append MR data to affiliations id needed
        if settings.PUBLISHER_SOCIETY_MR:
            j['affiliations'].append({
                'organization': {
                    'code': settings.PUBLISHER_SOCIETY_MR,
                    'name': SOCIETY_DICT.get(
                        settings.PUBLISHER_SOCIETY_MR,
                        ''
                    ).split(',')[0],
                },
                'affiliation_type': {
                    'code': 'MR',
                    'name': 'Mechanical Rights'
                },
                'territory': WORLD_DICT,
            })

        # append SR data to affiliations id needed
        if settings.PUBLISHER_SOCIETY_SR:
            j['affiliations'].append({
                'organization': {
                    'code': settings.PUBLISHER_SOCIETY_SR,
                    'name': SOCIETY_DICT.get(
                        settings.PUBLISHER_SOCIETY_SR,
                        ''
                    ).split(',')[0],
                },
                'affiliation_type': {
                    'code': 'SR',
                    'name': 'Synchronization Rights'
                },
                'territory': WORLD_DICT,
            })

        return j

    def get_dict(self, with_recordings=True):
        """Create a data structure that can be serialized as JSON.

        Normalize the structure if required.

        Returns:
            dict: JSON-serializable data structure
        """

        j = {
            'id': self.id,
            'code': self.work_id,
            'work_title': self.title,
            'last_change': self.last_change,
            'version_type': {
                'code': 'MOD',
                'name': 'Modified Version of a musical work',
            } if self.original_title else {
                'code': 'ORI',
                'name': 'Original Work',
            },
            'iswc': self.iswc,
            'other_titles': [
                at.get_dict() for at in self.alternatetitle_set.all()],

            'origin': (
                self.library_release.get_origin_dict() if self.library_release
                else None),
            'writers': [],
            'performing_artists': [],
            'original_works': [],
            'cross_references': []
        }

        if self.original_title:
            d = {'work_title': self.original_title}
            j['original_works'].append(d)

        # add data for (live) artists in work, normalize of required
        for aiw in self.artistinwork_set.all():
            d = aiw.get_dict()
            j['performing_artists'].append(d)

        # add data for writers in work, normalize of required
        for wiw in self.writerinwork_set.all():
            d = wiw.get_dict()
            j['writers'].append(d)

        if with_recordings:
            j['recordings'] = [
                recording.get_dict(with_releases=True)
                for recording in self.recordings.all()]

        # add cross references, currently only society work ids from ACKs
        used_society_codes = []
        for wa in self.workacknowledgement_set.all():
            if not wa.remote_work_id:
                continue
            if wa.society_code in used_society_codes:
                continue
            used_society_codes.append(wa.society_code)
            d = wa.get_dict()
            j['cross_references'].append(d)

        return j


class AlternateTitle(TitleBase):
    """Concrete class for alternate titles.

    Attributes:
        work (django.db.models.ForeignKey): Foreign key to Work model
        suffix (django.db.models.BooleanField): implies that the title should\
            be appended to the work title
    """

    work = models.ForeignKey(Work, on_delete=models.CASCADE)
    suffix = models.BooleanField(
        default=False,
        help_text='Select if this title is only a suffix to the main title.')

    class Meta:
        unique_together = (('work', 'title'),)
        ordering = ('-suffix', 'title')
        verbose_name = 'Alternative Title'

    def get_dict(self):
        """Create a data structure that can be serialized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """
        return {
            'title': str(self),
            'title_type': {
                'code': 'AT',
                'name': 'Alternative Title',
            },
        }

    def __str__(self):
        if self.suffix:
            return '{} {}'.format(self.work.title, self.title)
        return super().__str__()


class ArtistInWork(models.Model):
    """Artist performing the work (live in CWR 3).

    Attributes:
        artist (django.db.models.ForeignKey): FK to Artist
        work (django.db.models.ForeignKey): FK to Work
    """

    work = models.ForeignKey(Work, on_delete=models.CASCADE)
    artist = models.ForeignKey(Artist, on_delete=models.PROTECT)

    class Meta:
        verbose_name = 'Artist performing'
        verbose_name_plural = \
            'Artists performing (not mentioned in recordings section)'
        unique_together = (('work', 'artist'),)
        ordering = ('artist__last_name', 'artist__first_name')

    def __str__(self):
        return str(self.artist)

    def get_dict(self):
        """

        Returns:
            dict: taken from :meth:`models.Artist.get_dict`
        """
        return {'artist': self.artist.get_dict()}


class WriterInWork(models.Model):
    """Writers who created this work.

    At least one writer in work must be controlled.
    Sum of relative shares must be (roughly) 100%.
    Capacity is limited to roles for original writers.

    Attributes:
        work (django.db.models.ForeignKey): FK to Work
        writer (django.db.models.ForeignKey): FK to Writer
        saan (django.db.models.CharField): Society-assigned agreement number
            between the writer and the original publisher, please note that
            this field is for SPECIFIC agreements, for a general agreement,
            use :attr:`.base.IPIBase.saan`
        controlled (django.db.models.BooleanField): A complete mistery field
        relative_share (django.db.models.DecimalField): Initial split among
            writers, prior to publishing
        capacity (django.db.models.CharField): Role of the writer in this work
        publisher_fee (django.db.models.DecimalField): Percentage of royalties
            kept by publisher
    """

    class Meta:
        verbose_name = 'Writer in Work'
        verbose_name_plural = 'Writers in Work'
        unique_together = (('work', 'writer', 'controlled'),)
        ordering = (
            '-controlled', 'writer__last_name', 'writer__first_name', '-id')
    ROLES = (
        ('CA', 'Composer&Lyricist'),
        ('C ', 'Composer'),
        ('A ', 'Lyricist'),
        ('AR', 'Arranger'),
        ('AD', 'Adaptor'),
        ('TR', 'Translator'))

    work = models.ForeignKey(
        Work, on_delete=models.CASCADE)
    writer = models.ForeignKey(
        Writer, on_delete=models.PROTECT, blank=True, null=True)
    saan = models.CharField(
        'Society-assigned specific agreement number',
        help_text='Use this field for specific agreements only.'
                  'For general agreements use the field in the Writer form.',
        max_length=14, blank=True, null=True,
        validators=(CWRFieldValidator('saan'),), )
    controlled = models.BooleanField(default=False)
    relative_share = models.DecimalField(
        'Manuscript share', max_digits=5, decimal_places=2)
    capacity = models.CharField(
        'Role',
        max_length=2, blank=True, choices=ROLES)
    publisher_fee = models.DecimalField(
        max_digits=5, decimal_places=2, blank=True, null=True,
        validators=[MinValueValidator(0), MaxValueValidator(100)],
        help_text=(
            'Percentage of royalties kept by the publisher,\n'
            'in a specific agreement.'))

    def __str__(self):
        return str(self.writer)

    def clean_fields(self, *args, **kwargs):
        """Turn SAAN into uppercase.

        Args:
            *args: passing through
            **kwargs: passing through

        Returns:
            str: SAAN in uppercase
        """
        if self.saan:
            self.saan = self.saan.upper()
        return super().clean_fields(*args, **kwargs)

    def clean(self):
        """Make sure that controlled writers have all the required data.

        Also check that writers that are not controlled do not have data
        that can not apply to them."""

        if (self.writer and
                self.writer.generally_controlled and
                not self.controlled):
            raise ValidationError({
                'controlled': 'Must be set for a generally controlled writer.'
            })
        d = {}
        if self.controlled:
            if not self.capacity:
                d['capacity'] = 'Must be set for a controlled writer.'
            if not self.writer:
                d['writer'] = 'Must be set for a controlled writer.'
            else:
                if not self.writer._can_be_controlled:
                    d['writer'] = (
                        'IPI name and PR society must be set. '
                        'See "Writers" in the user manual')
        else:
            if self.saan:
                d['saan'] = 'Must be empty if writer is not controlled.'
            if self.publisher_fee:
                d['publisher_fee'] = \
                    'Must be empty if writer is not controlled.'
        if d:
            raise ValidationError(d)

    def get_agreement_dict(self):
        """Get agreement dictionary for this writer in work."""

        pub_pr_soc = settings.PUBLISHER_SOCIETY_PR
        pub_pr_name = SOCIETY_DICT.get(pub_pr_soc, '').split(',')[0]

        if not self.controlled or not self.writer:
            return None
        if self.writer.generally_controlled and not self.saan:
            # General
            return {
                'recipient_organization': {
                    'code': pub_pr_soc,
                    'name': pub_pr_name,
                },
                'recipient_agreement_number': self.writer.saan,
                'agreement_type': {
                    'code': 'OG', 'name': 'Original General',
                }
            }
        else:
            return {
                'recipient_organization': {
                    'code': pub_pr_soc,
                },
                'recipient_agreement_number': self.saan,
                'agreement_type': {
                    'code': 'OS',
                    'name': 'Original Specific',
                }
            }

    def get_dict(self):
        """Create a data structure that can be serialized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """

        j = {
            'writer': self.writer.get_dict() if self.writer else None,
            'controlled': self.controlled,
            'relative_share': str(self.relative_share / 100),
            'writer_role': {
                'code': self.capacity.strip(),
                'name': self.get_capacity_display()
            } if self.capacity else None,
            'original_publishers': [{
                'publisher': self.work.get_publisher_dict(),
                'publisher_role': {
                    'code': 'E',
                    'name': 'Original publisher'
                },
                'agreement': self.get_agreement_dict()
            }] if self.controlled else [],
        }
        return j


class Recording(models.Model):
    """Recording.

    Attributes:
        release_date (django.db.models.DateField): Recording Release Date
        duration (django.db.models.TimeField): Recording Duration
        isrc (django.db.models.CharField):
            International Standard Recording Code
        record_label (django.db.models.CharField): Record Label
    """

    class Meta:
        verbose_name = 'Recording'
        verbose_name_plural = 'Recordings'
        ordering = ('-id',)

    _recording_id = models.CharField(
        'Recording ID', max_length=14, blank=True, null=True, unique=True,
        editable=False,
        validators=(CWRFieldValidator('name'),))
    recording_title = models.CharField(
        blank=True, max_length=60,
        validators=(CWRFieldValidator('title'),))
    recording_title_suffix = models.BooleanField(
        default=False, help_text='A suffix to the WORK title.')
    version_title = models.CharField(
        blank=True, max_length=60,
        validators=(CWRFieldValidator('title'),))
    version_title_suffix = models.BooleanField(
        default=False, help_text='A suffix to the RECORDING title.')
    release_date = models.DateField(blank=True, null=True)
    duration = models.DurationField(blank=True, null=True)
    isrc = models.CharField(
        'ISRC', max_length=15, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('isrc'),))
    record_label = models.ForeignKey(
        Label, verbose_name='Record label',
        null=True, blank=True, on_delete=models.PROTECT)
    work = models.ForeignKey(
        Work, on_delete=models.CASCADE, related_name='recordings')
    artist = models.ForeignKey(
        Artist, verbose_name='Recording Artist', related_name='recordings',
        on_delete=models.PROTECT, blank=True, null=True)

    releases = models.ManyToManyField(Release, through='Track')

    audio_file = models.FileField(
        upload_to=upload_to, max_length=255, blank=True)

    def clean_fields(self, *args, **kwargs):
        """
        ISRC cleaning, just removing dots and dashes.

        Args:
            *args: may be used in upstream
            **kwargs: may be used in upstream

        Returns:
            return from :meth:`django.db.models.Model.clean_fields`

        """
        if not any([
            self.recording_title,
            self.version_title,
            self.release_date,
            self.isrc,
            self.duration,
            self.record_label,
            self.artist,
            self.audio_file]
        ):
            raise ValidationError('No data left, please delete instead.')
        if self.isrc:
            # Removing all characters added for readability
            self.isrc = self.isrc.replace('-', '').replace('.', '')
        return super().clean_fields(*args, **kwargs)

    @property
    def complete_recording_title(self):
        """
        Return complete recording title.

        Returns:
            str
        """
        if self.recording_title_suffix:
            return '{} {}'.format(
                self.work.title, self.recording_title).strip()
        return self.recording_title

    @property
    def complete_version_title(self):
        """
        Return complete version title.

        Returns:
            str
        """
        if self.version_title_suffix:
            return '{} {}'.format(
                self.complete_recording_title or self.work.title,
                self.version_title).strip()
        return self.version_title

    @property
    def title(self):
        """Generate title from various fields."""
        return (self.complete_version_title if self.version_title else
            self.complete_recording_title if self.recording_title else
            self.work.title) #  + (u' \U0001F508' if self.audio_file else '')

    @property
    def recording_id(self):
        """Create Recording ID used in registrations

        Returns:
            str: Internal Recording ID
        """
        if self._recording_id:
            return self._recording_id
        if self.id is None:
            return ''
        return '{}{:06}R'.format(settings.PUBLISHER_CODE, self.id)

    @recording_id.setter
    def recording_id(self, value):
        assert self._recording_id is None  # this should not be called if set
        if value:
            self._recording_id = value

    def __str__(self):
        """Return the most precise type of title"""
        if self.artist:
            return '{}: {} ({})'.format(
                self.recording_id, self.title, self.artist)
        else:
            return '{}: {}'.format(
                self.recording_id, self.title)

    def get_dict(self, with_releases=False, with_work=True):
        """Create a data structure that can be serialized as JSON.

        Args:
            with_releases (bool): add releases data (through tracks)
            with_work (bool): add work data

        Returns:
            dict: JSON-serializable data structure

        """
        j = {
            'id':
                self.id,
            'code':
                self.recording_id,
            'recording_title':
                self.complete_recording_title or self.work.title,
            'version_title':
                self.complete_version_title,
            'release_date':
                self.release_date.strftime('%Y%m%d') if self.release_date
                else None,
            'duration':
                duration_string(self.duration) if self.duration else None,
            'isrc':
                self.isrc,
            'recording_artist':
                self.artist.get_dict() if self.artist else None,
            'record_label':
                self.record_label.get_dict() if self.record_label else None,
        }
        if with_releases:
            j['tracks'] = []
            for track in self.tracks.all():
                d = track.release.get_dict()
                j['tracks'].append({
                    'release': d,
                    'cut_number': track.cut_number,
                })
        if with_work:
            j['works'] = [{'work': self.work.get_dict(with_recordings=False)}]
        return j


class Track(models.Model):
    """Track, a recording on a release.

    Attributes:
        recording (django.db.models.ForeignKey): Recording
        release (django.db.models.ForeignKey): Release
        cut_number (django.db.models.PositiveSmallIntegerField): Cut Number

    """

    class Meta:
        verbose_name = 'Track'
        unique_together = (('recording', 'release'), ('release', 'cut_number'))
        ordering = ('release', 'cut_number',)

    recording = models.ForeignKey(
        Recording, on_delete=models.PROTECT, related_name='tracks')
    release = models.ForeignKey(
        Release, on_delete=models.CASCADE, related_name='tracks')
    cut_number = models.PositiveSmallIntegerField(
        blank=True, null=True,
        validators=(MinValueValidator(1), MaxValueValidator(9999)))

    def get_dict(self):
        """Create a data structure that can be serialized as JSON.

        Returns:
            dict: JSON-serializable data structure

        """
        return {
            'cut_number': self.cut_number,
            'recording': self.recording.get_dict(
                with_releases=False, with_work=True)
        }
    
    def __str__(self):
        return self.recording.title


class DeferCwrManager(models.Manager):
    """Manager for CWR Exports and ACK Imports.

    Defers :attr:`CWRExport.cwr` and :attr:`AckImport.cwr` fields.

    """
    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.defer('cwr')
        return qs


class CWRExport(models.Model):
    """Export in CWR format.

    Common Works Registration format is a standard format for registration of
    musical works world-wide. Exports are available in CWR 2.1 revision 8 and
    CWR 3.0 (experimental).

    Attributes:
        nwr_rev (django.db.models.CharField): choice field where user can
            select which version and type of CWR it is
        cwr (django.db.models.TextField): contents of CWR file
        year (django.db.models.CharField): 2-digit year format
        num_in_year (django.db.models.PositiveSmallIntegerField): \
        CWR sequential number in a year
        works (django.db.models.ManyToManyField): included works
        description (django.db.models.CharField): internal note

    """

    class Meta:
        verbose_name = 'CWR Export'
        verbose_name_plural = 'CWR Exports'
        ordering = ('-id',)

    objects = DeferCwrManager()

    nwr_rev = models.CharField(
        'CWR version/type', max_length=3, db_index=True, default='NWR',
        choices=(
            ('NWR', 'CWR 2.1: New work registrations'),
            ('REV', 'CWR 2.1: Revisions of registered works'),
            ('NW2', 'CWR 2.2: New work registrations'),
            ('RE2', 'CWR 2.2: Revisions of registered works'),
            ('WRK', 'CWR 3.0: Work registration'),
            ('ISR', 'CWR 3.0: ISWC request (EDI)'),
            ('WR1', 'CWR 3.1 DRAFT: Work registration'),
        ))
    cwr = models.TextField(blank=True, editable=False)
    created_on = models.DateTimeField(editable=False, null=True)
    year = models.CharField(
        max_length=2, db_index=True, editable=False, blank=True)
    num_in_year = models.PositiveSmallIntegerField(default=0)
    works = models.ManyToManyField(Work, related_name='cwr_exports')
    description = models.CharField('Internal Note', blank=True, max_length=60)

    publisher_code = None
    agreement_pr = settings.PUBLISHING_AGREEMENT_PUBLISHER_PR
    agreement_mr = settings.PUBLISHING_AGREEMENT_PUBLISHER_MR
    agreement_sr = settings.PUBLISHING_AGREEMENT_PUBLISHER_SR

    @property
    def version(self):
        """Return CWR version."""
        if self.nwr_rev in ['WRK', 'ISR']:
            return '30'
        elif self.nwr_rev == 'WR1':
            return '31'
        elif self.nwr_rev in ['NW2', 'RE2']:
            return '22'
        return '21'

    @property
    def filename(self):
        """Return CWR file name.

        Returns:
            str: CWR file name
        """
        if self.version in ['30', '31']:
            return self.filename3
        return self.filename2

    @property
    def filename3(self):
        """Return proper CWR 3.x filename.

        Format is: CWYYnnnnSUB_REP_VM - m - r.EXT

        Returns:
            str: CWR file name
        """
        if self.version == '30':
            minor_version = '0-0'
        else:
            minor_version = '1-0'
        if self.nwr_rev == 'ISR':
            ext = 'ISR'
        else:
            ext = 'SUB'
        return 'CW{}{:04}{}_0000_V3-{}.{}'.format(
            self.year,
            self.num_in_year,
            self.publisher_code or settings.PUBLISHER_CODE,
            minor_version,
            ext)

    @property
    def filename2(self):
        """Return proper CWR 2.x filename.

        Returns:
            str: CWR file name
        """
        return 'CW{}{:04}{}_000.V{}'.format(
            self.year,
            self.num_in_year,
            self.publisher_code or settings.PUBLISHER_CODE,
            self.version)

    def __str__(self):
        return self.filename

    def get_record(self, key, record):
        """Create CWR record (row) from the key and dict.

        Args:
            key (str): type of record
            record (dict): field values

        Returns:
            str: CWR record (row)
        """
        if self.version == '30':
            template = TEMPLATES_30.get(key)
        elif self.version == '31':
            template = TEMPLATES_31.get(key)
        else:
            if self.version == '22':
                tdict = TEMPLATES_22
            else:
                tdict = TEMPLATES_21
            if (
                    key == 'HDR' and
                    len(record['ipi_name_number'].lstrip('0')) > 9
            ):
                # CWR 2.1 revision 8 "hack" for 10+ digit IPI name numbers
                template = tdict.get('HDR_8')
            else:
                template = tdict.get(key)
        record.update({'settings': settings})
        return template.render(Context(record)).upper()

    def get_transaction_record(self, key, record):
        """Create CWR transaction record (row) from the key and dict.

        This methods adds transaction and record sequences.

        Args:
            key (str): type of record
            record (dict): field values

        Returns:
            str: CWR record (row)
        """
        record['transaction_sequence'] = self.transaction_count
        record['record_sequence'] = self.record_sequence
        line = self.get_record(key, record)
        if line:
            self.record_count += 1
            self.record_sequence += 1
        return line

    def yield_iswc_request_lines(self, works):
        """Yield lines for an ISR (ISWC request) in CWR 3.x"""

        for work in works:

            # ISR
            self.record_sequence = 0
            if work['iswc']:
                work['indicator'] = 'U'
            yield self.get_transaction_record('ISR', work)

            # WRI
            reported = set()
            for wiw in work['writers']:
                w = wiw['writer']
                if not w:
                    continue  # goes to OWR
                tup = (w['code'], wiw['writer_role']['code'])
                if tup in reported:
                    continue
                reported.add(tup)
                w.update({
                    'writer_role': wiw['writer_role']['code'],
                })
                yield self.get_transaction_record('WRI', w)

            self.transaction_count += 1

    def yield_publisher_lines(self, publisher, controlled_relative_share):
        """Yield SPU/SPT lines.

        Args:
            controlled_relative_share (Decimal): sum of manuscript shares \
            for controlled writers

        Yields:
              str: CWR record (row/line)
        """
        affiliations = publisher.get('affiliations', [])
        for aff in affiliations:
            if aff['affiliation_type']['code'] == 'PR':
                publisher['pr_society'] = aff['organization']['code']
            elif aff['affiliation_type']['code'] == 'MR':
                publisher['mr_society'] = aff['organization']['code']
            elif aff['affiliation_type']['code'] == 'SR':
                publisher['sr_society'] = aff['organization']['code']

        pr_share = controlled_relative_share * self.agreement_pr
        mr_share = controlled_relative_share * self.agreement_mr
        sr_share = controlled_relative_share * self.agreement_sr
        yield self.get_transaction_record(
            'SPU', {
                'chain_sequence': 1,
                'name': publisher.get('name'),
                'code': publisher.get('code'),
                'ipi_name_number': publisher.get('ipi_name_number'),
                'ipi_base_number': publisher.get('ipi_base_number'),
                'pr_society': publisher.get('pr_society'),
                'mr_society': publisher.get('mr_society'),
                'sr_society': publisher.get('sr_society'),
                'pr_share': pr_share,
                'mr_share': mr_share,
                'sr_share': sr_share,
            })
        if controlled_relative_share:
            yield self.get_transaction_record(
                'SPT', {
                    'code': publisher.get('code'),
                    'pr_share': pr_share,
                    'mr_share': mr_share,
                    'sr_share': sr_share,
                    'pr_society': publisher.get('pr_society'),
                    'mr_society': publisher.get('mr_society'),
                    'sr_society': publisher.get('sr_society'),
                })

    def yield_registration_lines(self, works):
        """Yield lines for CWR registrations (WRK in 3.x, NWR and REV in 2.x)

        Args:
            works (list): list of work dicts

        Yields:
            str: CWR record (row/line)
        """
        for work in works:

            # WRK
            self.record_sequence = 0
            if self.version == '22':
                if self.nwr_rev == 'NW2':
                    record_type = 'NWR'
                elif self.nwr_rev == 'RE2':
                    record_type = 'REV'
            else:
                record_type = self.nwr_rev

            d = {
                'record_type': record_type,
                'code': work['code'],
                'work_title': work['work_title'],
                'iswc': work['iswc'],
                'recorded_indicator': 'Y' if work['recordings'] else 'U',
                'version_type': (
                    'MOD   UNSUNS'
                    if work['version_type']['code'] == 'MOD' else
                    'ORI         ')
            }
            yield self.get_transaction_record('WRK', d)
            yield from self.get_party_lines(work)
            yield from self.get_other_lines(work)
            self.transaction_count += 1

    def get_party_lines(self, work):
        """Yield SPU, SPT, OPU, SWR, SWT, OPT and PWR lines

        Args:
            work: musical work

        Yields:
            str: CWR record (row/line)
        """

        # SPU, SPT
        controlled_relative_share = Decimal(0)  # total pub share
        other_publisher_share = Decimal(0)  # used for co-publishing
        controlled_writer_ids = set()  # used for co-publishing
        copublished_writer_ids = set()  # used for co-publishing
        controlled_shares = defaultdict(Decimal)
        for wiw in work['writers']:
            if wiw['controlled']:
                controlled_writer_ids.add(wiw['writer']['code'])
        for wiw in work['writers']:
            writer = wiw['writer']
            share = Decimal(wiw['relative_share'])
            if wiw['controlled']:
                key = writer['code']
                controlled_relative_share += share
                controlled_shares[key] += share
            elif (writer and writer['code'] in controlled_writer_ids):
                key = writer['code']
                copublished_writer_ids.add(key)
                other_publisher_share += share
                controlled_shares[key] += share
        publisher = work['writers'][0]['original_publishers'][0]['publisher']
        yield from self.yield_publisher_lines(
            publisher, controlled_relative_share)
        # OPU, co-publishing only
        if other_publisher_share:
            pr_share = other_publisher_share * self.agreement_pr
            mr_share = other_publisher_share * self.agreement_mr
            sr_share = other_publisher_share * self.agreement_sr
            yield self.get_transaction_record(
                'OPU', {
                    'sequence': 2, 'pr_share': pr_share,
                    'mr_share': mr_share, 'sr_share': sr_share,
                })
            yield self.get_transaction_record(
                'OPT', {
                    'pr_share': pr_share, 'mr_share': mr_share, 
                    'sr_share': sr_share,})

        # SWR, SWT, PWR
        for wiw in work['writers']:
            if not wiw['controlled']:
                continue  # goes to OWR
            w = wiw['writer']
            agr = wiw['original_publishers'][0]['agreement']
            saan = agr['recipient_agreement_number'] if agr else None
            affiliations = w.get('affiliations', [])
            for aff in affiliations:
                if aff['affiliation_type']['code'] == 'PR':
                    w['pr_society'] = aff['organization']['code']
                elif aff['affiliation_type']['code'] == 'MR':
                    w['mr_society'] = aff['organization']['code']
                elif aff['affiliation_type']['code'] == 'SR':
                    w['sr_society'] = aff['organization']['code']
            share = controlled_shares[w['code']]
            pr_share = share * (1 - self.agreement_pr)
            mr_share = share * (1 - self.agreement_mr)
            sr_share = share * (1 - self.agreement_sr)
            w.update({
                'writer_role': wiw['writer_role']['code'],
                'share': share,
                'pr_share': pr_share,
                'mr_share': mr_share,
                'sr_share': sr_share,
                'saan': saan,
                'original_publishers': wiw['original_publishers']
            })
            yield self.get_transaction_record('SWR', w)
            if share:
                yield self.get_transaction_record('SWT', w)
            if share:
                yield self.get_transaction_record('MAN', w)
            w['publisher_sequence'] = 1
            w['publisher_code'] = publisher['code']
            w['publisher_name'] = publisher['name']
            yield self.get_transaction_record('PWR', w)
            if (self.version in ['30', '31'] and other_publisher_share and
                    w and w['code'] in copublished_writer_ids):
                w['publisher_sequence'] = 2
                yield self.get_transaction_record(
                    'PWR', {
                        'code': w['code'],
                        'publisher_sequence': 2
                    })

        # OWR
        for wiw in work['writers']:
            if wiw['controlled']:
                continue  # done in SWR
            writer = wiw['writer']
            if writer and writer['code'] in controlled_writer_ids:
                continue  # co-publishing, already solved
            if writer:
                w = wiw['writer']
                affiliations = w.get('affiliations', [])
                for aff in affiliations:
                    if aff['affiliation_type']['code'] == 'PR':
                        w['pr_society'] = aff['organization']['code']
                    elif aff['affiliation_type']['code'] == 'MR':
                        w['mr_society'] = aff['organization']['code']
                    elif aff['affiliation_type']['code'] == 'SR':
                        w['sr_society'] = aff['organization']['code']
            else:
                w = {'writer_unknown_indicator': 'Y'}
            share = Decimal(wiw['relative_share'])
            w.update({
                'writer_role': (wiw['writer_role']['code'] if wiw[
                    'writer_role'] else None),
                'share': share,
                'pr_share': share,
                'mr_share': share,
                'sr_share': share
            })
            yield self.get_transaction_record('OWR', w)
            if w['share']:
                yield self.get_transaction_record('OWT', w)
            if w['share']:
                yield self.get_transaction_record('MAN', w)
            if self.version in ['30', '31'] and other_publisher_share:
                w['publisher_sequence'] = 2
                yield self.get_transaction_record('PWR', w)

    def get_other_lines(self, work):
        """Yield ALT and subsequent lines

        Args:
            work: musical work

        Yields:
            str: CWR record (row/line)
        """

        # ALT
        alt_titles = set()
        for at in work['other_titles']:
            alt_titles.add(at['title'])
        for rec in work['recordings']:
            if rec['recording_title']:
                alt_titles.add(rec['recording_title'])
            if rec['version_title']:
                alt_titles.add(rec['version_title'])
        for alt_title in sorted(alt_titles):
            if alt_title == work['work_title']:
                continue
            yield self.get_transaction_record('ALT', {
                'alternate_title': alt_title
            })

        # VER
        if work['version_type']['code'] == 'MOD':
            yield self.get_transaction_record('OWK',
                                              work['original_works'][0])

        # PER
        # artists can be recording and/or live, so let's see
        artists = {}
        for aiw in work['performing_artists']:
            artists.update({aiw['artist']['code']: aiw['artist']})
        for rec in work['recordings']:
            if not rec['recording_artist']:
                continue
            artists.update({
                rec['recording_artist']['code']: rec['recording_artist']
            })
        for artist in artists.values():
            yield self.get_transaction_record('PER', artist)

        # REC
        for rec in work['recordings']:
            if rec['recording_artist']:
                rec['display_artist'] = '{} {}'.format(
                    rec['recording_artist']['first_name'] or '',
                    rec['recording_artist']['last_name'],
                ).strip()[:60]
            if rec['isrc']:
                rec['isrc_validity'] = 'Y'
            if rec['duration']:
                rec['duration'] = rec['duration'].replace(':', '')[0:6]
            if self.version in ['21', '22'] and not any([
                rec['release_date'],
                rec['duration'],
                rec['isrc'],
            ]):
                continue
            yield self.get_transaction_record('REC', rec)

        # ORN
        if work['origin']:
            yield self.get_transaction_record('ORN', {
                'library': work['origin']['library']['name'],
                'cd_identifier': work['origin']['cd_identifier'],
            })

        # XRF
        for xrf in work['cross_references']:
            yield self.get_transaction_record('XRF', xrf)

    def get_header(self):
        """Construct CWR HDR record."""
        return self.get_record('HDR', {
            'creation_date': datetime.now(),
            'filename': self.filename,
            'ipi_name_number': settings.PUBLISHER_IPI_NAME,
            'name': settings.PUBLISHER_NAME,
            'code': settings.PUBLISHER_CODE,
        })

    def yield_lines(self, works):
        """Yield CWR transaction records (rows/lines) for works

        Args:
            works (query): :class:`.models.Work` query

        Yields:
            str: CWR record (row/line)
        """

        self.record_count = self.record_sequence = self.transaction_count = 0

        yield self.get_header()

        if self.nwr_rev == 'NW2':
            yield self.get_record('GRH', {'transaction_type': 'NWR'})
        elif self.nwr_rev == 'RE2':
            yield self.get_record('GRH', {'transaction_type': 'REV'})
        else:
            yield self.get_record('GRH', {'transaction_type': self.nwr_rev})

        if self.nwr_rev == 'ISR':
            lines = self.yield_iswc_request_lines(works)
        else:
            lines = self.yield_registration_lines(works)

        for line in lines:
            yield line

        yield self.get_record('GRT', {
            'transaction_count': self.transaction_count,
            'record_count': self.record_count + 2
        })
        yield self.get_record('TRL', {
            'transaction_count': self.transaction_count,
            'record_count': self.record_count + 4
        })

    def create_cwr(self, publisher_code=None):
        """Create CWR and save.
        """
        now = timezone.now()
        if publisher_code is None:
            publisher_code = settings.PUBLISHER_CODE
        self.publisher_code = publisher_code
        if self.cwr:
            return
        self.created_on = now
        self.year = now.strftime('%y')
        nr = type(self).objects.filter(year=self.year)
        nr = nr.order_by('-num_in_year').first()
        if nr:
            self.num_in_year = nr.num_in_year + 1
        else:
            self.num_in_year = 1
        qs = self.works.order_by('id', )
        works = Work.objects.get_dict(qs)['works']
        self.cwr = ''.join(self.yield_lines(works))
        self.save()
        Work.persist_work_ids(self.works)


class WorkAcknowledgement(models.Model):
    """Acknowledgement of work registration.

    Attributes:
        date (django.db.models.DateField): Acknowledgement date
        remote_work_id (django.db.models.CharField): Remote work ID
        society_code (django.db.models.CharField): 3-digit society code
        status (django.db.models.CharField): 2-letter status code
        TRANSACTION_STATUS_CHOICES (tuple): choices for status
        work (django.db.models.ForeignKey): FK to Work
    """

    class Meta:
        verbose_name = 'Registration Acknowledgement'
        ordering = ('-date', '-id')
        index_together = (('society_code', 'remote_work_id'),)

    TRANSACTION_STATUS_CHOICES = (
        ('CO', 'Conflict'),
        ('DU', 'Duplicate'),
        ('RA', 'Transaction Accepted'),
        ('AS', 'Registration Accepted'),
        ('AC', 'Registration Accepted with Changes'),
        ('SR', 'Registration Accepted - Ready for Payment'),
        ('CR', 'Registration Accepted with Changes - Ready for Payment'),
        ('RJ', 'Rejected'),
        ('NP', 'No Participation'),
        ('RC', 'Claim rejected'),
        ('NA', 'Rejected - No Society Agreement Number'),
        ('WA', 'Rejected - Wrong Society Agreement Number'),
    )

    work = models.ForeignKey(Work, on_delete=models.PROTECT)
    society_code = models.CharField(
        'Society', max_length=3, choices=SOCIETIES)
    date = models.DateField()
    status = models.CharField(max_length=2, choices=TRANSACTION_STATUS_CHOICES)
    remote_work_id = models.CharField(
        'Remote work ID', max_length=20, blank=True, db_index=True)

    def get_dict(self):
        """
        Return dictionary with external work IDs.

        Returns:
            dict: JSON-serializable data structure
        """
        # if not self.remote_work_id:
        #     return None
        j = {
            'organization': {
                'code': self.society_code,
                'name': self.get_society_code_display().split(',')[0],
            },
            'identifier': self.remote_work_id,
        }
        return j


class ACKImport(models.Model):
    """CWR acknowledgement file import.

    Attributes:
        filename (django.db.models.CharField): Description
        society_code (models.CharField): 3-digit society code,
            please note that ``choices`` is not set.
        society_name (models.CharField): Society name,
            used if society code is missing.
        date (django.db.models.DateField): Acknowledgement date
        report (django.db.models.CharField): Basically a log
        cwr (django.db.models.TextField): contents of CWR file
    """

    class Meta:
        verbose_name = 'CWR ACK Import'
        ordering = ('-date', '-id')

    objects = DeferCwrManager()

    filename = models.CharField(max_length=60, editable=False)
    society_code = models.CharField(max_length=3, editable=False)
    society_name = models.CharField(max_length=45, editable=False)
    date = models.DateField(editable=False)
    report = models.TextField(editable=False)
    cwr = models.TextField(blank=True, editable=False)

    def __str__(self):
        return self.filename


class DataImport(models.Model):
    """Importing basic work data from a CSV file.

    This class just acts as log, the actual logic is in :mod:`.data_import`.
    """

    class Meta:
        verbose_name = 'Data Import'
        ordering = ('-date', '-id')

    filename = models.CharField(max_length=60, editable=False)
    report = models.TextField(editable=False)
    date = models.DateTimeField(auto_now_add=True, editable=False)

    def __str__(self):
        return self.filename


def smart_str_conversion(value):
    """Convert to Title Case only if UPPER CASE."""
    if value.isupper():
        return value.title()
    return value


FORCE_CASE_CHOICES = {
    'upper': str.upper,
    'title': str.title,
    'smart': smart_str_conversion,
}


@receiver(pre_save)
def change_case(sender, instance, **kwargs):
    """Change case of CharFields from :mod:`music_publisher`."""
    force_case = FORCE_CASE_CHOICES.get(settings.OPTION_FORCE_CASE)
    if not force_case:
        return
    if sender._meta.app_label != 'music_publisher':
        return
    for field in instance._meta.get_fields():
        if isinstance(field, models.CharField):
            value = getattr(instance, field.name)
            if (isinstance(value, str) and
                    field.editable and
                    field.choices is None and (
                        'name' in field.name or
                        'title' in field.name)):
                value = force_case(value)
                setattr(instance, field.name, value)



"""
Main interface for :mod:`music_publisher`.

All views are here, except for :mod:`.royalty_calculation`.

"""
import base64
import re
import uuid
import zipfile
from csv import DictWriter
from datetime import datetime
from decimal import Decimal

from django import forms
from django.conf import settings
from django.contrib import admin, messages
from django.core.exceptions import ValidationError
from django.db import models
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.html import mark_safe
from django.utils.timezone import now

from .forms import (
    ACKImportForm, AlternateTitleFormSet, DataImportForm, LibraryReleaseForm,
    WorkForm, WriterInWorkFormSet, PlaylistForm
)
from .models import (
    ACKImport, AlternateTitle, Artist, ArtistInWork, CWRExport, Playlist,
    CommercialRelease, DataImport, Label, Library, LibraryRelease, Recording,
    Release, SOCIETY_DICT, Track, Work, WorkAcknowledgement, Writer,
    WriterInWork,
)
from .validators import CWRFieldValidator

IS_POPUP_VAR = admin.options.IS_POPUP_VAR


class ImageWidget(forms.widgets.ClearableFileInput):
    template_name = 'admin/widgets/image.html'


class AudioPlayerWidget(forms.widgets.ClearableFileInput):
    template_name = 'admin/widgets/audioplayer.html'


class MusicPublisherAdmin(admin.ModelAdmin):
    """Parent class to all admin classes."""
    save_as = True


class ArtistInWorkInline(admin.TabularInline):
    """Inline interface for :class:`.models.ArtistInWork`.
    """
    autocomplete_fields = ('artist', 'work')
    model = ArtistInWork
    extra = 0
    ordering = ('artist__last_name', 'artist__first_name')
    verbose_name_plural = \
        'Artists performing Works (not mentioned in "recordings" section)'


class RecordingInline(admin.StackedInline):
    """Inline interface for :class:`.models.Recording`,
        used in :class:`WorkAdmin`.
    """
    autocomplete_fields = ('artist', 'work', 'record_label')
    readonly_fields = ('complete_recording_title', 'complete_version_title')
    show_change_link = True

    def get_fieldsets(self, request, obj=None):
        if settings.OPTION_FILES:
            return (
                ('Metadata', {
                    'fields': (
                        'work',
                        ('recording_title', 'recording_title_suffix',
                         'complete_recording_title'),
                        ('version_title', 'version_title_suffix',
                         'complete_version_title'),
                        ('isrc', 'record_label', 'artist'),
                        ('duration', 'release_date'),
                    ),
                }),
                ('Audio', {
                    'fields': (
                        ('audio_file',)
                    ),
                }),
            )
        else:
            return (
                (None, {
                    'fields': (
                        'work',
                        ('recording_title', 'recording_title_suffix',
                         'complete_recording_title'),
                        ('version_title', 'version_title_suffix',
                         'complete_version_title'),
                        ('isrc', 'record_label', 'artist'),
                        ('duration', 'release_date'),
                    ),
                }),
            )

    formfield_overrides = {
        models.FileField: {'widget': AudioPlayerWidget},
        models.TimeField: {'widget': forms.TimeInput},
    }

    verbose_name_plural = \
        'Recordings (with recording artists and record labels)'
    model = Recording
    ordering = ('recording_title', 'version_title', 'id')
    extra = 0


@admin.register(Artist)
class ArtistAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Artist`.
    """

    ordering = ('last_name', 'first_name', 'isni', '-id')

    list_display = (
        'last_or_band', 'first_name', 'isni', 'recording_count', 'work_count')
    search_fields = ('last_name', 'isni',)

    def get_fieldsets(self, request, obj=None):
        if settings.OPTION_FILES:
            return (
                ('Name', {'fields': (('first_name', 'last_name'),)}),
                ('ISNI', {'fields': ('isni',), }),
                ('Public', {'fields': ('image', 'description')}),
                ('Internal', {'fields': ('notes',), }),
            )
        else:
            return (
                ('Name', {'fields': (('first_name', 'last_name'),)}),
                ('ISNI', {'fields': ('isni',), }),
                ('Internal', {'fields': ('notes',), }),
            )

    formfield_overrides = {
        models.ImageField: {'widget': ImageWidget},
    }

    def last_or_band(self, obj):
        """Placeholder for :attr:`.models.Artist.last_name`."""
        return obj.last_name

    last_or_band.short_description = 'Last or band name'
    last_or_band.admin_order_field = 'last_name'

    actions = None

    def save_model(self, request, obj, form, *args, **kwargs):
        """Save, then update ``last_change`` of the works whose CWR
        registration changes due to this change.
        """
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(
                models.Q(artistinwork__artist=obj) |
                models.Q(recordings__artist=obj))
            qs.update(last_change=now())

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(models.Count('work', distinct=True))
        qs = qs.annotate(recording__count=models.Count(
            'recordings', distinct=True))
        return qs

    def work_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.work__count

        url = reverse('admin:music_publisher_work_changelist')
        url += '?artists__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    work_count.short_description = 'Perf. Works'
    work_count.admin_order_field = 'work__count'

    def recording_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.recording__count

        url = reverse('admin:music_publisher_recording_changelist')
        url += '?artist__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    recording_count.short_description = 'Recordings'
    recording_count.admin_order_field = 'recording__count'


@admin.register(Label)
class LabelAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Label`.
    """
    actions = None
    search_fields = ('name',)
    list_display = (
        'name', 'recording_count', 'commercialrelease_count',
        'libraryrelease_count')
    readonly_fields = (
        'recording_count', 'commercialrelease_count', 'libraryrelease_count')

    formfield_overrides = {
        models.ImageField: {'widget': ImageWidget},
    }

    def get_fieldsets(self, request, obj=None):
        if settings.OPTION_FILES:
            return (
                ('Name', {'fields': ('name',)}),
                ('Public', {'fields': ('image', 'description')}),
                ('Internal', {'fields': ('notes',)}),
            )
        else:
            return (
                ('Name', {'fields': ('name',)}),
                ('Notes', {'fields': ('notes',), }),
            )

    ordering = ('name', '-id')

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(
            libraryrelease__count=models.Count(
                'release', distinct=True,
                filter=models.Q(release__cd_identifier__isnull=False)))
        qs = qs.annotate(
            commercialrelease__count=models.Count(
                'release', distinct=True,
                filter=models.Q(release__cd_identifier__isnull=True)))
        qs = qs.annotate(models.Count('recording', distinct=True))
        return qs

    def commercialrelease_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.commercialrelease__count

        url = reverse('admin:music_publisher_commercialrelease_changelist')
        url += '?release_label__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    commercialrelease_count.short_description = 'Commercial releases'
    commercialrelease_count.admin_order_field = 'commercialrelease__count'

    def libraryrelease_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.libraryrelease__count

        url = reverse('admin:music_publisher_libraryrelease_changelist')
        url += '?release_label__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    libraryrelease_count.short_description = 'Library releases'
    libraryrelease_count.admin_order_field = 'libraryrelease__count'

    def recording_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.recording__count

        url = reverse('admin:music_publisher_recording_changelist')
        url += '?record_label__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    recording_count.short_description = 'Recordings'
    recording_count.admin_order_field = 'recording__count'

    def save_model(self, request, obj, form, *args, **kwargs):
        """Save, then update ``last_change`` of the corresponding works.
        """
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(
                models.Q(recordings__record_label=obj))
            qs.update(last_change=now())


@admin.register(Library)
class LibraryAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Library`.
    """
    actions = None
    search_fields = ('name',)
    ordering = ('name', '-id')

    list_display = ('name', 'libraryrelease_count', 'work_count')
    readonly_fields = ('libraryrelease_count', 'work_count')
    fields = ('name',)

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(
            work__count=models.Count('release__works', distinct=True))
        qs = qs.annotate(
            release__count=models.Count('release__id', distinct=True))
        return qs

    def libraryrelease_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.release__count

        url = reverse('admin:music_publisher_libraryrelease_changelist')
        url += '?library__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    libraryrelease_count.short_description = 'Library releases'
    libraryrelease_count.admin_order_field = 'libraryrelease__count'

    def work_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.work__count

        url = reverse('admin:music_publisher_work_changelist')
        url += '?library_release__library__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    work_count.short_description = 'Works'
    work_count.admin_order_field = 'work__count'

    def save_model(self, request, obj, form, *args, **kwargs):
        """Save, then update ``last_change`` of the corresponding works.
        """
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(
                models.Q(library_release__library=obj))
            qs.update(last_change=now())


class TrackInline(admin.TabularInline):
    """Inline interface for :class:`.models.Track`, used in
    :class:`LibraryReleaseAdmin` and :class:`CommercialReleaseAdmin`.
    """
    model = Track
    ordering = ('release', 'cut_number',)
    autocomplete_fields = ('release', 'recording')
    extra = 0


class PlaylistTrackInline(TrackInline):
    def has_audio(self, obj):
        return bool(obj.recording.audio_file)
    has_audio.boolean = True
    fields = ('recording', 'has_audio')
    readonly_fields = ('has_audio',)


@admin.register(Release)
class ReleaseAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Release`.
    """

    ordering = ('release_title', 'cd_identifier', '-id')
    actions = None
    list_display = (
        '__str__',
    )

    formfield_overrides = {
        models.ImageField: {'widget': ImageWidget},
    }

    search_fields = ('release_title', '^cd_identifier')

    def has_module_permission(self, request):
        """Return False"""
        return False

    def has_add_permission(self, request):
        """Return False"""
        return False

    def has_change_permission(self, request, obj=None):
        """Return False"""
        return False

    def has_delete_permission(self, request, obj=None):
        """Return False"""
        return False


@admin.register(LibraryRelease)
class LibraryReleaseAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.LibraryRelease`.
    """

    ordering = ('release_title', 'cd_identifier', '-id')
    form = LibraryReleaseForm
    inlines = [TrackInline]
    autocomplete_fields = ('release_label', 'library')

    fieldsets = (
        ('Library', {
            'fields': (
                ('library', 'cd_identifier'),)
        }),
        ('Release (album) metadata', {
            'fields': (
                ('release_title', 'release_label'),
                ('ean', 'release_date'))
        }),
        ('Public', {
            'fields': ('image', 'description')
        }),
    )

    list_display = (
        'cd_identifier',
        'library',
        'release_title',
        'release_label',
        'release_date',
        'work_count',
        'track_count',
    )
    readonly_fields = ('work_count', 'track_count')

    list_filter = ('release_label', 'library')
    search_fields = ('release_title', '^cd_identifier')

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return []
        return super().get_inline_instances(request)

    def save_model(self, request, obj, form, *args, **kwargs):
        """Save, then update ``last_change`` of the corresponding works.
        """
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(library_release=obj)
            qs.update(last_change=now())

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(models.Count('tracks', distinct=True))
        qs = qs.annotate(models.Count('works', distinct=True))
        return qs

    def work_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.works__count

        url = reverse('admin:music_publisher_work_changelist')
        url += '?library_release__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    work_count.short_description = 'Works'
    work_count.admin_order_field = 'works__count'

    def track_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.tracks__count

        url = reverse('admin:music_publisher_recording_changelist')
        url += '?release__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    track_count.short_description = 'Recordings'
    track_count.admin_order_field = 'tracks__count'

    def create_json(self, request, qs):
        """Batch action that downloads a JSON file containing library releases.

        Returns:
            JsonResponse: JSON file with selected works
        """

        j = LibraryRelease.objects.get_dict(qs)

        response = JsonResponse(j, json_dumps_params={'indent': 4})
        name = '{}-libraryreleases-{}'.format(
            settings.PUBLISHER_CODE, datetime.now().toordinal())
        cd = 'attachment; filename="{}.json"'.format(name)
        response['Content-Disposition'] = cd
        return response

    create_json.short_description = \
        'Export selected library releases (JSON).'

    actions = ['create_json']

    def get_actions(self, request):
        """Custom action disabling the default ``delete_selected``."""
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions


@admin.register(Playlist)
class PlaylistAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Playlist`.
    """

    ordering = ('-id',)
    form = PlaylistForm
    inlines = [PlaylistTrackInline]

    fieldsets = (
        (None, {
            'fields': (
                ('release_title', 'release_date'),
                'description', 'image'
            )
        }),
        ('URLs', {
            'fields': (
                'secret_url',
                'secret_api_url',
            )
        }),
    )

    # autocomplete_fields = ['recordings']

    list_display = (
        'release_title',
        'valid',
        'secret_url',
        'track_count',
    )

    def valid(self, obj):
        if obj.id is None:
            return False
        if obj.release_date and obj.release_date < now().date():
            return False
        return True
    valid.boolean = True
    readonly_fields = ('cd_identifier', 'secret_url', 'secret_api_url')

    search_fields = ('release_title', '^cd_identifier')

    def secret_url(self, obj):
        if self.valid(obj):
            url = self.request.build_absolute_uri(obj.secret_url)
            return mark_safe(f'<a href="{ url }" target="_blank">{ url }</a>')
        return ''
    secret_url.short_description = 'Secret URL'

    def secret_api_url(self, obj):
        if self.valid(obj):
            url = self.request.build_absolute_uri(obj.secret_api_url)
            return mark_safe(f'<a href="{ url }" target="_blank">{ url }</a>')
        return ''
    secret_api_url.short_description = 'Secret API URL'

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return []
        return super().get_inline_instances(request)

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        self.request = request
        qs = super().get_queryset(request)
        qs = qs.exclude(release_date__gt=now())
        qs = qs.annotate(models.Count('tracks', distinct=True))
        return qs

    def track_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.tracks__count

        url = reverse('admin:music_publisher_recording_changelist')
        url += '?release__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    track_count.short_description = 'Recordings'
    track_count.admin_order_field = 'tracks__count'

    # def create_json(self, request, qs):
    #     """Batch action that downloads a JSON file containing library releases.
    # 
    #     Returns:
    #         JsonResponse: JSON file with selected works
    #     """
    # 
    #     j = Playlist.objects.get_dict(qs)
    # 
    #     response = JsonResponse(j, json_dumps_params={'indent': 4})
    #     name = '{}-libraryreleases-{}'.format(
    #         settings.PUBLISHER_CODE, datetime.now().toordinal())
    #     cd = 'attachment; filename="{}.json"'.format(name)
    #     response['Content-Disposition'] = cd
    #     return response
    # 
    # create_json.short_description = \
    #     'Export selected library releases (JSON).'
    # 
    # actions = ['create_json']


@admin.register(CommercialRelease)
class CommercialReleaseAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.CommercialRelease`.
    """

    ordering = ('release_title', 'cd_identifier', '-id')
    inlines = [TrackInline]
    autocomplete_fields = ('release_label',)

    list_display = (
        'release_title',
        'release_label',
        'release_date',
        'track_count',
    )

    readonly_fields = ('track_count',)

    list_filter = ('release_label',)
    search_fields = ('release_title',)

    fieldsets = (
        ('Release (album) metadata', {
            'fields': (
                ('release_title', 'release_label'),
                ('ean', 'release_date'))
        }),
        ('Public', {
            'fields': ('image', 'description')
        }),
    )

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return []
        return super().get_inline_instances(request)

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(models.Count('tracks', distinct=True))
        return qs

    def track_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.tracks__count

        url = reverse('admin:music_publisher_recording_changelist')
        url += '?release__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    track_count.short_description = 'Recordings'
    track_count.admin_order_field = 'tracks__count'

    def create_json(self, request, qs):
        """Batch action that downloads a JSON file containing commercial
        releases.

        Returns:
            JsonResponse: JSON file with selected commercial releases
        """

        j = CommercialRelease.objects.get_dict(qs)

        response = JsonResponse(j, json_dumps_params={'indent': 4})
        name = '{}-libraryreleases-{}'.format(
            settings.PUBLISHER_CODE, datetime.now().toordinal())
        cd = 'attachment; filename="{}.json"'.format(name)
        response['Content-Disposition'] = cd
        return response

    create_json.short_description = \
        'Export selected commercial releases (JSON).'

    actions = ['create_json']

    def get_actions(self, request):
        """Custom action disabling the default ``delete_selected``."""
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions


@admin.register(Writer)
class WriterAdmin(MusicPublisherAdmin):
    """Interface for :class:`.models.Writer`.
    """

    ordering = ('last_name', 'first_name', 'ipi_name', '-id')
    list_display = ('last_name', 'first_name', 'ipi_name', 'pr_society',
                    '_can_be_controlled', 'generally_controlled',
                    'work_count')

    list_filter = ('_can_be_controlled', 'generally_controlled', 'pr_society')
    search_fields = ('last_name', 'ipi_name')
    readonly_fields = ('_can_be_controlled', 'work_count')

    formfield_overrides = {
        models.ImageField: {'widget': ImageWidget},
    }

    def get_fieldsets(self, request, obj=None):
        """Return the fieldsets.

        Depending on settings, MR and PR affiliations may not be needed.
        See :meth:`WriterAdmin.get_society_list`"""
        if settings.OPTION_FILES:
            return [
                ('Name', {
                    'fields': (
                        ('first_name', 'last_name'),)
                }),
                ('IPI', {
                    'fields': (
                        ('ipi_name', 'ipi_base'),),
                }),
                ('Societies', {
                    'fields': (
                        self.get_society_list(),),
                }),
                ('General agreement', {
                    'fields': (
                        ('generally_controlled',
                         ('saan', 'publisher_fee'))
                    ),
                }),
                ('Public', {'fields': ('image', 'description')}),
                ('Internal', {
                    'fields': ('notes',),
                }),
            ]
        else:
            return [
                ('Name', {
                    'fields': (
                        ('first_name', 'last_name'),)
                }),
                ('IPI', {
                    'fields': (
                        ('ipi_name', 'ipi_base'),),
                }),
                ('Societies', {
                    'fields': (
                        self.get_society_list(),),
                }),
                ('General agreement', {
                    'fields': (
                        ('generally_controlled',
                         ('saan', 'publisher_fee'))
                    ),
                }),
                ('Internal', {
                    'fields': ('notes',),
                }),
            ]

    actions = None

    @staticmethod
    def get_society_list():
        """List which society fields are required.

        Mechanical and Sync affiliation is not required if writers don't
        collect any of it, which is the most usual case."""

        societies = ['pr_society']
        if settings.PUBLISHING_AGREEMENT_PUBLISHER_MR != Decimal(1):
            societies.append('mr_society')
        if settings.PUBLISHING_AGREEMENT_PUBLISHER_SR != Decimal(1):
            societies.append('sr_society')
        return societies

    def save_model(self, request, obj, form, *args, **kwargs):
        """Perform normal save_model, then update last_change of
        all connected works."""
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(writerinwork__writer=obj)
            qs.update(last_change=now())

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(work__count=models.Count('works', distinct=True))
        return qs

    def work_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.work__count

        url = reverse('admin:music_publisher_work_changelist')
        url += '?writers__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    work_count.short_description = 'Works'
    work_count.admin_order_field = 'work__count'


class AlternateTitleInline(admin.TabularInline):
    """Inline interface for :class:`.models.AlternateTitle`.
    """
    model = AlternateTitle
    formset = AlternateTitleFormSet
    extra = 0
    readonly_fields = ('complete_alt_title',)
    verbose_name_plural = \
        'Alternative titles (not mentioned in "recordings" section)'
    fields = ('title', 'suffix', 'complete_alt_title')
    ordering = ('suffix', 'title',)

    def complete_alt_title(self, obj):
        """Return the complete title, see
        :meth:`.models.AlternateTitle.__str__`"""
        return str(obj)


class WriterInWorkInline(admin.TabularInline):
    """Inline interface for :class:`.models.WriterInWork`.
    """

    autocomplete_fields = ('writer',)
    model = WriterInWork
    formset = WriterInWorkFormSet
    extra = 0
    min_num = 1  # One writer is required
    fields = (
        'writer', 'capacity', 'relative_share', 'controlled', 'saan',
        'publisher_fee')
    ordering = (
            '-controlled', 'writer__last_name', 'writer__first_name', '-id')


class WorkAcknowledgementInline(admin.TabularInline):
    """Inline interface for :class:`.models.WorkAcknowledgement`,
        used in :class:`WorkAdmin`.

        Note that normal users should only have a 'view' permission.
    """

    model = WorkAcknowledgement
    extra = 0
    fields = ('date', 'society_code', 'remote_work_id', 'status')
    ordering = ('-date', '-id')


@admin.register(Work)
class WorkAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Work`.

    This is by far the most important part of the interface.

    Attributes:
        actions (tuple): batch actions used:
            :meth:`create_cwr`,
            :meth:`create_json`
        inlines (tuple): inlines used in change view:
            :class:`AlternateTitleInline`,
            :class:`WriterInWorkInline`,
            :class:`RecordingInline`,
            :class:`ArtistInWorkInline`,
            :class:`WorkAcknowledgementInline`,
    """

    ordering = ('-id',)
    form = WorkForm

    inlines = (
        WriterInWorkInline,
        RecordingInline,
        AlternateTitleInline,
        ArtistInWorkInline,
        WorkAcknowledgementInline)

    def writer_last_names(self, obj):
        """This is a standard way how writers are shown in other apps."""
        return obj.writer_last_names()

    writer_last_names.short_description = 'Writers\' last names'
    writer_last_names.admin_order_field = 'writers__last_name'

    def percentage_controlled(self, obj):
        """Controlled percentage
        (sum of relative shares for controlled writers)

        Please note that writers in work are already included in the queryset
        for other reasons, so no overhead except summing.
        """
        return sum(
            wiw.relative_share for wiw in obj.writerinwork_set.all()
            if wiw.controlled)

    percentage_controlled.short_description = '% controlled'

    def work_id(self, obj):
        """Return :attr:`.models.Work.work_id`, make it sortable."""
        return obj.work_id

    work_id.short_description = 'Work ID'
    work_id.admin_order_field = 'id'

    def cwr_export_count(self, obj):
        """Return the count of CWR exports with the link to the filtered
        changelist view for :class:`CWRExportAdmin`."""

        count = obj.cwr_exports__count
        url = reverse('admin:music_publisher_cwrexport_changelist')
        url += '?works__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    cwr_export_count.short_description = 'CWRs'
    cwr_export_count.admin_order_field = 'cwr_exports__count'

    def recording_count(self, obj):
        """Return the count of CWR exports with the link to the filtered
        changelist view for :class:`CWRExportAdmin`."""

        count = obj.recordings__count
        url = reverse('admin:music_publisher_recording_changelist')
        url += '?work__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    recording_count.short_description = 'Recordings'
    recording_count.admin_order_field = 'recordings__count'

    readonly_fields = (
        'writer_last_names', 'work_id', 'cwr_export_count')
    list_display = (
        'work_id', 'title', 'iswc', 'writer_last_names',
        'percentage_controlled', 'library_release',
        'recording_count', 'cwr_export_count')

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.prefetch_related('library_release__library')
        qs = qs.prefetch_related('writerinwork_set__writer')
        qs = qs.annotate(models.Count('cwr_exports', distinct=True))
        qs = qs.annotate(models.Count('recordings', distinct=True))
        return qs

    class InCWRListFilter(admin.SimpleListFilter):
        """Custom list filter if work is included in any of CWR files.
        """

        title = 'In CWR'
        parameter_name = 'in_cwr'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter if in any of CWR files.
            """
            if self.value() == 'Y':
                return queryset.exclude(cwr_exports__count=0)
            elif self.value() == 'N':
                return queryset.filter(cwr_exports__count=0)

    class ACKSocietyListFilter(admin.SimpleListFilter):
        """Custom list filter of societies from ACK files.
        """

        title = 'Acknowledgement society'
        parameter_name = 'ack_society'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """

            codes = WorkAcknowledgement.objects.order_by()
            codes = codes.values_list(
                'society_code', flat=True).distinct()
            return sorted(
                [(code, SOCIETY_DICT.get(code, code)) for code in codes],
                key=lambda code: code[1])

        def queryset(self, request, queryset):
            """Filter on society sending ACKs.
            """
            if self.value():
                queryset = queryset.filter(
                    workacknowledgement__society_code=self.value()).distinct()
                queryset.society_code = self.value()
            return queryset

    class ACKStatusListFilter(admin.SimpleListFilter):
        """Custom list filter on ACK status.
        """

        title = 'Acknowledgement status'
        parameter_name = 'ack_status'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return WorkAcknowledgement.TRANSACTION_STATUS_CHOICES

        def queryset(self, request, queryset):
            """Filter on ACK status.
            """
            if self.value():
                if hasattr(queryset, 'society_code'):
                    queryset = queryset.filter(
                        workacknowledgement__status=self.value(),
                        workacknowledgement__society_code=queryset.society_code
                    ).distinct()
                else:
                    queryset = queryset.filter(
                        workacknowledgement__status=self.value()).distinct()
            return queryset

    class HasISWCListFilter(admin.SimpleListFilter):
        """Custom list filter on the presence of ISWC.
        """

        title = 'Has ISWC'
        parameter_name = 'has_iswc'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter on presence of :attr:`.iswc`.
            """
            if self.value() == 'Y':
                return queryset.exclude(iswc__isnull=True)
            elif self.value() == 'N':
                return queryset.filter(iswc__isnull=True)

    class HasRecordingListFilter(admin.SimpleListFilter):
        """Custom list filter on the presence of recordings.
        """
        title = 'Has Recordings'
        parameter_name = 'has_rec'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter on presence of :class:`.models.Recording`.
            """
            if self.value() == 'Y':
                return queryset.exclude(recordings__count=0)
            elif self.value() == 'N':
                return queryset.filter(recordings__count=0)

    list_filter = (
        HasISWCListFilter,
        HasRecordingListFilter,
        ('library_release__library', admin.RelatedOnlyFieldListFilter),
        ('library_release', admin.RelatedOnlyFieldListFilter),
        ('writers', admin.RelatedOnlyFieldListFilter),
        'last_change',
        InCWRListFilter,
        ACKSocietyListFilter,
        ACKStatusListFilter,
    )

    search_fields = (
        'title', 'alternatetitle__title', '^iswc', '^id', '^_work_id',
        'recordings__recording_title', 'recordings__version_title',
        '^recordings__isrc', 'writerinwork__writer__last_name')

    def get_search_results(self, request, queryset, search_term):
        """Deal with the situation term is work ID.
        """
        if search_term.isnumeric():
            search_term = search_term.lstrip('0')
        return super().get_search_results(request, queryset, search_term)

    fieldsets = (
        (None, {
            'fields': (
                'work_id',
                ('title', 'iswc'),
                ('original_title', 'version_type')
            )
        }),
        ('Library (Production music only)', {
            'fields': (
                ('library_release',),
            )
        }),
    )

    autocomplete_fields = ('library_release',)

    def save_model(self, request, obj, form, *args, **kwargs):
        """Set last_change if the work form has changed."""
        if form.changed_data:
            obj.last_change = now()
        super().save_model(request, obj, form, *args, **kwargs)

    def save_formset(self, request, form, formset, change):
        """Set last_change for the work if any of the inline forms has changed.
        """
        save_instance = False
        for form in formset:
            if form.changed_data:
                save_instance = True
        super().save_formset(request, form, formset, change)
        if save_instance:
            formset.instance.last_change = now()
            formset.instance.save()

    def create_cwr(self, request, qs):
        """Batch action that redirects to the add view for
        :class:`CWRExportAdmin` with selected works.
        """
        url = reverse('admin:music_publisher_cwrexport_add')
        work_ids = qs.values_list('id', flat=True)
        view = CWRExportAdmin(CWRExport, admin.site).add_view(
            request, url, work_ids=work_ids)
        return view

    create_cwr.short_description = 'Create CWR from selected works.'

    def create_json(self, request, qs):
        """Batch action that downloads a JSON file containing selected works.

        Returns:
            JsonResponse: JSON file with selected works
        """

        Work.persist_work_ids(qs)

        j = Work.objects.get_dict(qs)
        response = JsonResponse(j, json_dumps_params={'indent': 4})
        name = '{}-works-{}'.format(
            settings.PUBLISHER_CODE, datetime.now().toordinal())
        cd = 'attachment; filename="{}.json"'.format(name)
        response['Content-Disposition'] = cd
        return response

    create_json.short_description = \
        'Export selected works (JSON).'

    def get_labels_for_csv(self, works, repeating_column_nr=0, simple=False):
        """Return the list of labels for the CSV file."""
        labels = [
            'Work ID',
            'Work Title', 'ISWC', 'Original Title', 'Library', 'CD Identifier',
        ]
        alt_title_max = repeating_column_nr
        writer_max = repeating_column_nr
        writer_with_publisher_max = repeating_column_nr
        artist_max = repeating_column_nr
        xrf_max = repeating_column_nr
        recording_max = repeating_column_nr
        for work in works:
            alt_title_max = max(alt_title_max, len(work.get('other_titles')))
            writer_max = max(writer_max, len(work.get('writers')))
            ops = 0
            for w in work.get('writers'):
                if w.get('original_publishers'):
                    ops += 1
            writer_with_publisher_max = max(writer_with_publisher_max, ops)
            recording_max = max(recording_max, len(work.get('recordings')))
            artist_max = max(artist_max, len(work.get('performing_artists')))
            xrf_max = max(xrf_max, len(work.get('cross_references')))
        for i in range(alt_title_max):
            labels.append('Alt Title {}'.format(i + 1))
        for i in range(writer_max):
            labels.append('Writer {} Last'.format(i + 1))
            labels.append('Writer {} First'.format(i + 1))
            labels.append('Writer {} IPI'.format(i + 1))
            labels.append('Writer {} PRO'.format(i + 1))
            if not simple:
                labels.append('Writer {} MRO'.format(i + 1))
                labels.append('Writer {} SRO'.format(i + 1))
            labels.append('Writer {} Role'.format(i + 1))
            labels.append('Writer {} Manuscript Share'.format(i + 1))
            if not simple:
                labels.append('Writer {} PR Share'.format(i + 1))
                labels.append('Writer {} MR Share'.format(i + 1))
                labels.append('Writer {} SR Share'.format(i + 1))
            labels.append('Writer {} Controlled'.format(i + 1))
            if i < writer_with_publisher_max:
                labels.append('Writer {} SAAN'.format(i + 1))
            if not simple and i < writer_with_publisher_max:
                labels.append('Writer {} Publisher Name'.format(i + 1))
                labels.append('Writer {} Publisher IPI'.format(i + 1))
                labels.append('Writer {} Publisher PRO'.format(i + 1))
                labels.append('Writer {} Publisher MRO'.format(i + 1))
                labels.append('Writer {} Publisher SRO'.format(i + 1))
                labels.append('Writer {} Publisher PR Share'.format(i + 1))
                labels.append('Writer {} Publisher MR Share'.format(i + 1))
                labels.append('Writer {} Publisher SR Share'.format(i + 1))
        for i in range(recording_max):
            if not simple:
                labels.append('Recording {} ID'.format(i + 1))
                labels.append('Recording {} Recording Title'.format(i + 1))
                labels.append('Recording {} Version Title'.format(i + 1))
            labels.append('Recording {} Release Date'.format(i + 1))
            labels.append('Recording {} Duration'.format(i + 1))
            labels.append('Recording {} ISRC'.format(i + 1))
            if not simple:
                labels.append('Recording {} Artist Last'.format(i + 1))
                labels.append('Recording {} Artist First'.format(i + 1))
                labels.append('Recording {} Artist ISNI'.format(i + 1))
                labels.append('Recording {} Record Label'.format(i + 1))
        for i in range(artist_max):
            labels.append('Artist {} Last'.format(i + 1))
            labels.append('Artist {} First'.format(i + 1))
            labels.append('Artist {} ISNI'.format(i + 1))
        if not simple:
            for i in range(xrf_max):
                labels.append('Reference {} CMO'.format(i + 1))
                labels.append('Reference {} ID'.format(i + 1))
        return labels

    def get_rows_for_csv(self, works):
        """Return rows for the CSV file, including the header."""

        class EchoWriter:
            """Class with write() method just echoing values."""
            def write(self, value):
                return value

        PR = settings.PUBLISHING_AGREEMENT_PUBLISHER_PR
        MR = settings.PUBLISHING_AGREEMENT_PUBLISHER_MR
        SR = settings.PUBLISHING_AGREEMENT_PUBLISHER_SR

        pseudo_buffer = EchoWriter()
        labels = self.get_labels_for_csv(works)
        writer = DictWriter(pseudo_buffer, labels)
        header = dict(zip(labels, labels))
        yield writer.writerow(header)
        # yield writer.writeheader()  # In Python 3.8
        for work in works:
            ows = work.get('original_works')
            row = {
                'Work ID': work['code'],
                'Work Title': work['work_title'],
                'ISWC': work.get('iswc', ''),
            }
            if ows:
                row['Original Title'] = ows[0]['work_title']
            origin = work.get('origin')
            if origin:
                row['Library'] = origin['library']['name']
                row['CD Identifier'] = origin['cd_identifier']
            for i, alt in enumerate(work['other_titles']):
                row['Alt Title {}'.format(i + 1)] = alt['title']
            for i, wiw in enumerate(work['writers']):
                w = wiw.get('writer') or {}
                row['Writer {} Last'.format(i + 1)] = w.get('last_name', '')
                row['Writer {} First'.format(i + 1)] = w.get('first_name', '')
                row['Writer {} IPI'.format(i + 1)] = w.get(
                    'ipi_name_number', '')
                role = wiw.get('writer_role', {})
                if role:
                    row['Writer {} Role'.format(i + 1)] = '{} - {}'.format(
                        role['code'], role['name'])
                for aff in w.get('affiliations', []):
                    code = aff['affiliation_type']['code']
                    cmo = aff['organization']
                    row['Writer {} {}O'.format(i + 1, code)] = \
                        '{} - {}'.format(cmo['code'], cmo['name'])
                ops = wiw.get('original_publishers')

                row['Writer {} Manuscript Share'.format(i + 1)] = Decimal(
                    wiw.get('relative_share', '0')
                ).quantize(Decimal('0.0001'))
                if ops:
                    op = ops[0]
                    agreement = op.get('agreement')
                    saan = agreement.get('recipient_agreement_number', '')
                    row['Writer {} SAAN'.format(i + 1)] = saan
                    agreement_type = agreement['agreement_type']['code']
                    if agreement_type == 'OG':
                        controlled = 'General Agreement'
                    else:
                        controlled = 'Yes'
                    row['Writer {} Publisher Name'.format(
                        i + 1)] = op['publisher']['name']
                    row['Writer {} Publisher IPI'.format(
                        i + 1)] = op['publisher']['ipi_name_number']
                    for aff in op['publisher'].get('affiliations', []):
                        row['Writer {} Publisher {}O'.format(
                            i + 1,
                            aff['affiliation_type']['code']
                        )] = '{} - {}'.format(
                            aff['organization']['code'],
                            aff['organization']['name'])
                    row['Writer {} PR Share'.format(i + 1)] = (
                        Decimal( wiw.get('relative_share')) * (1 - PR)
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} Publisher PR Share'.format(i + 1)] = (
                        Decimal(wiw.get('relative_share')) * PR
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} MR Share'.format(i + 1)] = (
                        Decimal(wiw.get('relative_share')) * (1 - MR)
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} Publisher MR Share'.format(i + 1)] = (
                            Decimal(wiw.get('relative_share')) * MR
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} SR Share'.format(i + 1)] = (
                        Decimal(wiw.get('relative_share')) * (1 - SR)
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} Publisher SR Share'.format(i + 1)] = (
                        Decimal(wiw.get('relative_share')) * SR
                    ).quantize(Decimal('0.0001'))
                else:
                    controlled = 'No'
                    row['Writer {} PR Share'.format(i + 1)] = Decimal(
                        wiw.get('relative_share', '0')
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} MR Share'.format(i + 1)] = Decimal(
                        wiw.get('relative_share', '0')
                    ).quantize(Decimal('0.0001'))
                    row['Writer {} SR Share'.format(i + 1)] = Decimal(
                        wiw.get('relative_share', '0')
                    ).quantize(Decimal('0.0001'))
                row['Writer {} Controlled'.format(i + 1)] = controlled
            for i, rec in enumerate(work['recordings']):
                row['Recording {} ID'.format(i + 1)] = rec['code']
                row['Recording {} Recording Title'.format(i + 1)] = \
                    rec['recording_title']
                row['Recording {} Version Title'.format(i + 1)] = \
                    rec['version_title']
                row['Recording {} Release Date'.format(i + 1)] = \
                    rec['release_date']
                row['Recording {} Duration'.format(i + 1)] = rec['duration']
                row['Recording {} ISRC'.format(i + 1)] = rec['isrc']
                row['Recording {} Record Label'.format(i + 1)] = \
                    (rec['record_label'] or {}).get('name')
                artist = rec.get('recording_artist') or {}
                row['Recording {} Artist Last'.format(i + 1)] = \
                    artist.get('last_name', '')
                row['Recording {} Artist Last'.format(i + 1)] = \
                    artist.get('last_name', '')
                row['Recording {} Artist First'.format(i + 1)] = \
                    artist.get('first_name', '')
                row['Recording {} Artist ISNI'.format(i + 1)] = \
                    artist.get('isni', '')
            for i, aiw in enumerate(work['performing_artists']):
                artist = aiw.get('artist')
                row['Artist {} Last'.format(i + 1)] = \
                    artist.get('last_name', '')
                row['Artist {} First'.format(i + 1)] = \
                    artist.get('first_name', '')
                row['Artist {} ISNI'.format(i + 1)] = \
                    artist.get('isni', '')
            for i, xrf in enumerate(work['cross_references']):
                code = xrf['organization']['code']
                name = xrf['organization']['name']
                row['Reference {} CMO'.format(i + 1)] = '{} - {}'.format(
                    code, name)
                row['Reference {} ID'.format(i + 1)] = xrf['identifier']
            yield writer.writerow(row)

    def create_csv(self, request, qs):
        """Batch action that downloads a CSV file containing selected works.

        Returns:
            JsonResponse: JSON file with selected works
        """

        Work.persist_work_ids(qs)

        j = Work.objects.get_dict(qs)
        works = j.get('works', [])
        response = HttpResponse(
            self.get_rows_for_csv(works),
            content_type="text/csv")
        name = '{}{}'.format(
            settings.PUBLISHER_CODE, datetime.now().toordinal())
        cd = 'attachment; filename="{}.csv"'.format(name)
        response['Content-Disposition'] = cd
        return response

    create_csv.short_description = \
        'Export selected works (CSV).'

    actions = (create_cwr, create_json, create_csv)

    def get_actions(self, request):
        """Custom action disabling the default ``delete_selected``."""
        actions = super().get_actions(request)
        if not settings.PUBLISHER_CODE:
            del actions['create_cwr']
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        instances = super().get_inline_instances(request)
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return [i for i in instances if type(i) not in [
                RecordingInline, WorkAcknowledgementInline
            ]]
        return instances


@admin.register(Recording)
class RecordingAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Recording`."""

    actions = None
    list_display = (
        'recording_id', 'title', 'isrc', 'has_audio', 
        'work_link', 'artist_link', 'label_link')
    ordering = ('-id',)

    formfield_overrides = {
        models.FileField: {'widget': AudioPlayerWidget},
        models.TimeField: {'widget': forms.TimeInput},
    }

    def has_audio(self, obj):
        return bool(obj.audio_file)
    has_audio.boolean = True

    class HasISRCListFilter(admin.SimpleListFilter):
        """Custom list filter on the presence of ISRC.
        """

        title = 'Has ISRC'
        parameter_name = 'has_isrc'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter on presence of :attr:`.iswc`.
            """
            if self.value() == 'Y':
                return queryset.exclude(isrc__isnull=True)
            elif self.value() == 'N':
                return queryset.filter(isrc__isnull=True)

    class HasAudioFilter(admin.SimpleListFilter):
        """Custom list filter on the presence of audio file.
        """

        title = 'Has audio'
        parameter_name = 'has_audio_file'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter on presence of :attr:`.iswc`.
            """
            if self.value() == 'Y':
                return queryset.exclude(audio_file='')
            elif self.value() == 'N':
                return queryset.filter(audio_file='')

    list_filter = (HasISRCListFilter, HasAudioFilter, 'artist', 'record_label')

    def lookup_allowed(self, lookup, value):
        allowed = super().lookup_allowed(lookup, value)
        return allowed

    search_fields = (
        'work__title', 'recording_title', 'version_title', 'isrc')
    autocomplete_fields = ('artist', 'work', 'record_label')
    readonly_fields = (
        'recording_id',
        'complete_recording_title', 'complete_version_title', 'title',
        'has_audio',
        'work_link', 'artist_link', 'label_link')

    def get_fieldsets(self, request, obj=None):
        if settings.OPTION_FILES:
            return (
                ('Metadata', {
                    'fields': (
                        'recording_id',
                        'work',
                        ('recording_title', 'recording_title_suffix',
                         'complete_recording_title'),
                        ('version_title', 'version_title_suffix',
                         'complete_version_title'),
                        ('isrc', 'record_label', 'artist'),
                        ('duration', 'release_date'),
                    ),
                }),
                ('Audio', {
                    'fields': ('audio_file',),
                }),
            )
        else:
            return (
                (None, {
                    'fields': (
                        'recording_id',
                        'work',
                        ('recording_title', 'recording_title_suffix',
                         'complete_recording_title'),
                        ('version_title', 'version_title_suffix',
                         'complete_version_title'),
                        ('isrc', 'record_label', 'artist'),
                        ('duration', 'release_date'),
                    ),
                }),
            )

    def get_queryset(self, request):
        """Optimized query regarding work name
        """
        qs = super().get_queryset(request)
        qs = qs.prefetch_related('work__writers')
        qs = qs.prefetch_related('artist')
        qs = qs.prefetch_related('record_label')
        return qs

    def recording_id(self, obj):
        """Return :attr:`.models.Recording.recording_id`, make it sortable."""
        return obj.recording_id

    recording_id.short_description = 'Recording ID'
    recording_id.admin_order_field = 'id'

    def title(self, obj):
        """Return the recording title, which is not the necessarily the
        title field."""
        return obj.title

    def work_link(self, obj):
        """Link to the work the recording is based on."""
        url = reverse('admin:music_publisher_work_change', args=[obj.work.id])
        link = '<a href="{}">{}</a>'.format(url, obj.work)
        return mark_safe(link)

    work_link.short_description = 'Work'
    work_link.admin_order_field = 'work__id'

    def artist_link(self, obj):
        """Link to the recording artist."""
        if not obj.artist:
            return None
        url = reverse(
            'admin:music_publisher_artist_change', args=[obj.artist.id])
        link = '<a href="{}">{}</a>'.format(url, obj.artist)
        return mark_safe(link)

    artist_link.short_description = 'Recording Artist'
    artist_link.admin_order_field = 'artist'

    def label_link(self, obj):
        """Link to the recording label."""
        if not obj.record_label:
            return None
        url = reverse(
            'admin:music_publisher_label_change', args=[obj.record_label.id])
        link = '<a href="{}">{}</a>'.format(url, obj.record_label)
        return mark_safe(link)

    label_link.short_description = 'Record Label'
    label_link.admin_order_field = 'record_label'


@admin.register(CWRExport)
class CWRExportAdmin(admin.ModelAdmin):
    """Admin interface for :class:`.models.CWRExport`.
    """

    actions = None
    ordering = ('-id',)

    def work_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        count = obj.works__count

        url = reverse('admin:music_publisher_work_changelist')
        url += '?cwr_exports__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))

    work_count.short_description = 'Works'
    work_count.admin_order_field = 'works__count'

    def get_preview(self, obj):
        """Get CWR preview.

        If you are using highlighing, then override this method."""

        return obj.cwr or ''

    def view_link(self, obj):
        """Link to the CWR preview."""
        if obj.created_on:
            url = reverse(
                'admin:music_publisher_cwrexport_change', args=(obj.id,))
            url += '?preview=true'
            return mark_safe(
                '<a href="{}" target="_blank">View CWR</a>'.format(url))

    def download_link(self, obj):
        """Link for downloading CWR file."""
        if obj.created_on:
            url = reverse(
                'admin:music_publisher_cwrexport_change', args=(obj.id,))
            url += '?download=true'
            return mark_safe('<a href="{}">Download</a>'.format(url))

    def get_queryset(self, request):
        """Optimized query with count of works in the export.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(models.Count('works', distinct=True))
        return qs

    def date(self, obj):
        if obj and obj.created_on:
            return obj.created_on.date()

    autocomplete_fields = ('works',)
    list_display = (
        'filename', 'nwr_rev', 'date', 'work_count', 'view_link',
        'download_link', 'description')
    list_editable = ('description',)

    list_filter = ('nwr_rev', 'year')
    search_fields = ('description', 'works__title', 'num_in_year')

    def get_readonly_fields(self, request, obj=None):
        """Read-only fields differ if CWR has been completed."""
        if obj and obj.cwr:
            return (
                'nwr_rev', 'description', 'works', 'filename', 'view_link',
                'download_link')
        else:
            return ()

    def get_fields(self, request, obj=None):
        """Shown fields differ if CWR has been completed."""
        if obj and obj.cwr:
            return (
                'nwr_rev', 'description', 'works', 'filename', 'view_link',
                'download_link')
        else:
            return ('nwr_rev', 'description', 'works')

    def has_add_permission(self, request):
        """Return false if CWR delivery code is not present."""
        if not settings.PUBLISHER_CODE:
            return False
        return super().has_add_permission(request)

    def has_delete_permission(self, request, obj=None):
        """If CWR has been created, it can no longer be deleted, as it may
        have been sent. This may change once the delivery is automated."""
        if obj and obj.cwr:
            return False
        return super().has_delete_permission(request, obj)

    def has_change_permission(self, request, obj=None):
        """If object exists, it can only be edited in changelist."""
        if not settings.PUBLISHER_CODE:
            return False
        if obj:
            return False
        return super().has_delete_permission(request, obj)

    def get_form(self, request, obj=None, **kwargs):
        """Set initial values for work IDs."""
        form = super().get_form(request, obj, **kwargs)
        if hasattr(self, 'work_ids'):
            form.base_fields['works'].initial = self.work_ids
        return form

    def add_view(self, request, form_url='', extra_context=None,
                 work_ids=None):
        """Added work_ids as default for wizard from
        :meth:`WorkAdmin.create_cwr`."""
        if work_ids:
            self.work_ids = work_ids
            request.method = 'GET'
        return super().add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        """Normal change view with two sub-views defined by GET parameters:

        Parameters:
            preview: that returns the preview of CWR file,
            download: that downloads the CWR file."""
        obj = get_object_or_404(CWRExport, pk=object_id)
        if 'preview' in request.GET:
            cwr = self.get_preview(obj)
            return render(request, 'raw_cwr.html', {
                **self.admin_site.each_context(request),
                'version': obj.version,
                'lines': cwr.split('\r\n'),
                'title': obj.filename
            })
        elif 'download' in request.GET:
            response = HttpResponse(content_type='application/zip')
            zip_file = zipfile.ZipFile(
                response, 'w', zipfile.ZIP_DEFLATED)
            zip_file.writestr(obj.filename, obj.cwr.encode().decode('latin1'))
            if obj.version in ['30', '31']:
                cd = 'attachment; filename="{}.zip"'.format(
                    obj.filename.replace('.', '_'))
            else:
                cd = 'attachment; filename="{}"'.format(
                    obj.filename.replace(
                        '.V21', '.zip'
                    ).replace('.V22', '.zip'))
            response['Content-Disposition'] = cd
            return response

        extra_context = {
            'show_save': False,
        }
        if obj.cwr:
            extra_context.update({
                'save_as': False,
                'show_save_and_continue': False,
                'show_delete': False,
            })
        return super().change_view(
            request, object_id, form_url='', extra_context=extra_context)

    def save_related(self, request, form, formsets, change):
        """:meth:`save_model` passes the main object, which is needed to fetch
            CWR from the external service, but only after related objects are
            saved.
        """
        super().save_related(request, form, formsets, change)
        form.instance.create_cwr()


class AdminWithReport(admin.ModelAdmin):
    """The parent class for all admin classes with a report field."""
    def print_report(self, obj):
        """Mark report as HTML-safe."""
        return mark_safe(obj.report)
    print_report.short_description = 'Report'
    ordering = ('-id',)


@admin.register(ACKImport)
class ACKImportAdmin(AdminWithReport):
    """Admin interface for :class:`.models.ACKImport`.
    """

    def get_form(self, request, obj=None, **kwargs):
        """Returns a custom form for new objects, default one for changes.
        """
        if obj is None:
            return ACKImportForm
        return super().get_form(request, obj, **kwargs)

    list_display = (
        'filename', 'society_code', 'society_name', 'date', 'view_link')
    list_filter = ('society_code', 'society_name')
    fields = readonly_fields = (
        'filename', 'society_code', 'society_name', 'date', 'print_report',
        'view_link')

    add_fields = ('acknowledgement_file', 'import_iswcs')

    def get_fields(self, request, obj=None):
        """Return different fields for add vs change.
        """
        if obj:
            return self.fields
        return self.add_fields

    RE_ACK_21 = re.compile(
        r'(?<=\n)ACK.{43}(NWR|REV).{60}(.{20})(.{20})(.{8})(.{2})(.*?)(?=^ACK|^GRT)', re.S | re.M)
    RE_ACK_30 = re.compile(
        r'(?<=\n)ACK.{43}(WRK).{60}(.{20})(.{20}){20}(.{8})(.{2})(.*?)(?=^ACK|^GRT)', re.S | re.M)
    RE_ISW_21 = re.compile(
        r'(?<=\n)ISW.{78}(.{14})(.{11}).*?(?=^ISW|^GRT)', re.S | re.M)

    def process(self, request, ack_import, file_content, import_iswcs=False):
        """Create appropriate WorkAcknowledgement objects, without duplicates.

        Big part of this code should be moved to the model, left here because
        messaging is simpler.
        """

        society_code = ack_import.society_code
        ack_import_url = reverse(
            'admin:music_publisher_ackimport_change', 
            args=(ack_import.id,))
        ack_import_link = f'<a href="{ack_import_url}">{ack_import}</a>'
        from django.contrib.admin.models import CHANGE, LogEntry

        if import_iswcs:
            validator = CWRFieldValidator('iswc')

        unknown_work_ids = []
        existing_work_ids = []
        report = ''
        if file_content[59:64] == '01.10':
            pattern = self.RE_ACK_21
        else:
            pattern = self.RE_ACK_30
        for x in re.findall(pattern, file_content):
            tt, work_id, remote_work_id, dat, status, rest = x
            if import_iswcs and rest:
                header = rest.strip().split('\n')[0]
                iswc = header[95:106].strip() or None
                if iswc:
                    try:
                        validator(iswc)
                    except ValidationError:
                        iswc = None
            # work ID is numeric with an optional string
            work_id = work_id.strip()
            remote_work_id = remote_work_id.strip()
            dat = datetime.strptime(dat, '%Y%m%d').date()
            work = Work.objects.filter(_work_id=work_id).first()
            if not work:
                unknown_work_ids.append(work_id)
                continue
            if import_iswcs and iswc:
                if work.iswc:
                    if work.iswc != iswc:
                        report +=(
                            'A different ISWC exists for work ' +
                            '{}: {} (old) vs {} (new).<br/>\n'.format(
                                work, work.iswc, iswc) +
                            'Old ISWC kept, please investigate.<br/>\n')
                        self.message_user(
                            request,
                            'Conflicting ISWCs found for work {}!'.format(
                                work),
                            level=messages.ERROR
                        )
                else:
                    duplicate = Work.objects.exclude(id=work.id)
                    duplicate = duplicate.filter(iswc__iexact=iswc).first()
                    if duplicate:
                        report +=(
                            'One ISWC can not be used for two works: ' +
                            '{} {} {}.<br/>\n'.format(iswc, duplicate, work) +
                            'This usually happens if one work is entered '
                            'twice. ' +
                            'ISWC not imported for {}.<br/>\n'.format(work))
                        self.message_user(
                            request,
                            'Duplicate works found for ISWC {}!'.format(iswc),
                            level=messages.ERROR
                        )
                    else:
                        work.iswc = iswc
                        work.last_change = now()
                        s = f'ISWC imported from ACK file: {ack_import_link}.'
                        LogEntry.objects.log_action(
                            request.user.id,
                            admin.options.get_content_type_for_model(work).id,
                            work.id, str(work), CHANGE, s)
                        work.save()
            wa, c = WorkAcknowledgement.objects.get_or_create(
                work_id=work.id,
                remote_work_id=remote_work_id,
                society_code=society_code,
                date=dat,
                status=status)
            if not c:
                existing_work_ids.append(str(work_id))
                continue
            url = reverse(
                'admin:music_publisher_work_change', args=(work.id,))
            report += '<a href="{}">{}</a> {} &mdash; {}<br/>\n'.format(
                url, work.work_id, work.title, wa.get_status_display())
        if file_content[59:64] == '01.10':
            for work_id, iswc in re.findall(self.RE_ISW_21, file_content):
                work_id = work_id.strip()
                work = Work.objects.filter(_work_id=work_id).first()
                if not work:
                    unknown_work_ids.append(work_id)
                    continue
                if import_iswcs and iswc:
                    if work.iswc:
                        if work.iswc != iswc:
                            report += (
                                    'A different ISWC exists for work ' +
                                    '{}: {} (old) vs {} (new).<br/>\n'.format(
                                        work, work.iswc, iswc) +
                                    'Old ISWC kept, please '
                                    'investigate.<br/>\n')
                            self.message_user(
                                request,
                                'Conflicting ISWCs found for work {}!'.format(
                                    work),
                                level=messages.ERROR
                            )
                    else:
                        work.iswc = iswc
                        work.last_change = now()
                        s = f'ISWC imported from ISW file: {ack_import_link}.'
                        LogEntry.objects.log_action(
                            request.user.id,
                            admin.options.get_content_type_for_model(work).id,
                            work.id, str(work), CHANGE, s)
                        work.save()
        if unknown_work_ids:
            messages.add_message(
                request, messages.ERROR,
                'Unknown work IDs: {}'.format(', '.join(unknown_work_ids)))
        if existing_work_ids:
            messages.add_message(
                request, messages.ERROR,
                'Data already exists for some or all works. '
                'Affected work IDs: {}'.format(', '.join(existing_work_ids)))
        return report

    def save_model(self, request, obj, form, change):
        """Custom save_model, it ignores changes, validates the form for new
            instances, if valid, it processes the file and, upon success,
            calls ``super().save_model``."""

        if form.is_valid():
            cd = form.cleaned_data
            obj.filename = cd['filename']
            obj.society_code = cd['society_code']
            obj.society_name = cd['society_name']
            obj.date = cd['date']
            # TODO move process() to model, and handle messages here
            super().save_model(request, obj, form, change)
            obj.report = self.process(
                request, obj, cd['acknowledgement_file'],
                cd['import_iswcs'])
            obj.cwr = cd['acknowledgement_file']
            super().save_model(request, obj, form, True)

    def has_add_permission(self, request):
        """Return false if CWR delivery code is not present."""
        if not settings.PUBLISHER_CODE:
            return False
        return super().has_add_permission(request)

    def has_delete_permission(self, request, obj=None, *args, **kwargs):
        """Deleting ACK imports is a really bad idea.
        """
        return False

    def has_change_permission(self, request, obj=None):
        """Deleting this would make no sense, since the data is processed."""
        return False

    def get_preview(self, obj):
        """Get CWR preview.

        If you are using highlighing, then override this method."""

        return obj.cwr or ''

    def view_link(self, obj):
        """Link to CWR ACK preview."""
        url = reverse(
            'admin:music_publisher_ackimport_change', args=(obj.id,))
        url += '?preview=true'
        return mark_safe(
            '<a href="{}" target="_blank">View CWR</a>'.format(url))

    def change_view(self, request, object_id, form_url='', extra_context=None):
        """Normal change view with a sub-view defined by GET parameters:

        Parameters:
            preview: that returns the preview of CWR file."""
        try:
            obj = get_object_or_404(ACKImport, pk=int(object_id))
        except ValueError:
            return super().change_view(
                request, object_id, form_url='', extra_context=extra_context)
        if 'preview' in request.GET:
            cwr = self.get_preview(obj)
            if (cwr[59:64] == '01.10'):
                version = '21'
            else:
                version = '30'  # never seen one yet
            try:
                return render(request, 'raw_cwr.html', {
                    **self.admin_site.each_context(request),
                    'version': version,
                    'lines': cwr.split('\n'),
                    'title': obj.filename
                })
            except Exception:  # Parsing user garbage, could be anything
                return render(request, 'raw_cwr.html', {
                    **self.admin_site.each_context(request),
                    'version': '',
                    'lines': cwr.split('\n'),
                    'title': obj.filename
                })

        return super().change_view(
            request, object_id, form_url='', extra_context=extra_context)


@admin.register(DataImport)
class DataImportAdmin(AdminWithReport):
    """Data import from CSV files.

    Only the interface is here, the whole logic is in
    :mod:`.data_import`.
    """

    add_form_template = 'admin/add_data_import.html'
    form = DataImportForm

    list_display = ('filename', 'date')
    fields = readonly_fields = ('filename', 'date', 'print_report')
    ordering = ('-id',)

    def add_view(self, request, form_url='', extra_context=None):
        if 'download_template' in request.GET:
            fieldnames = WorkAdmin.get_labels_for_csv(None, [], 6, simple=True)
            response = HttpResponse(
                ','.join(fieldnames),
                content_type="text/csv")
            cd = 'attachment; filename="{}"'.format(
                'DMP_data_exchange_template.csv')
            response['Content-Disposition'] = cd
            return response
        return super().add_view(request, form_url, extra_context)

    add_fields = ('data_file', 'ignore_unknown_columns')

    def get_fields(self, request, obj=None):
        """Return different fields for add vs change.
        """
        if obj:
            return self.fields
        return self.add_fields

    def has_delete_permission(self, request, obj=None, *args, **kwargs):
        """Deleting data imports is a really bad idea.
        """
        return False

    def has_change_permission(self, request, obj=None):
        """Deleting this would make no sense, since the data is processed."""
        return False

    def get_form(self, request, obj=None, change=False, **kwargs):
        form = super().get_form(request, obj, change, **kwargs)
        form.user = request.user
        return form

    def save_model(self, request, obj, form, change):
        """Custom save_model, it ignores changes, validates the form for new
            instances, if valid, it processes the file and, upon success,
            calls ``super().save_model``."""

        if form.is_valid():
            cd = form.cleaned_data
            f = cd['data_file']
            obj.filename = f.name
            obj.report = cd['report']
            super().save_model(request, obj, form, change)



"""Filter used in DMP dashboard."""

from django import template
from collections import OrderedDict

register = template.Library()


def yield_sections(model_dict, sections):
    """Convert model dictionary according to section structure"""
    for name, object_names in sections.items():
        models = []
        for object_name in object_names:
            if object_name in model_dict:
                models.append(model_dict[object_name])
                del model_dict[object_name]
        if models:
            yield {
                'name': name,
                'models': models}


@register.filter(name='dmp_model_groups')
def dmp_model_groups(model_list):
    """Return groups of models."""
    model_dict = OrderedDict([(el['object_name'], el) for el in model_list])
    sections = {
        'Musical Works': [
            'Work', 'Publisher', 'Writer', 'CWRExport', 'ACKImport',
            'DataImport', 'RoyaltyCalculation'],
        'Recordings': ['Recording', 'Artist', 'Label', 'Playlist'],
        'Releases': [
            'CommercialRelease', 'LibraryRelease', 'Library'],
    }
    yield from yield_sections(model_dict, sections)



import setuptools

with open('README.rst', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-music-publisher',
    version='22.1',
    author='Matija Kolarić',
    author_email='matijakolaric@users.noreply.github.com',
    license='MIT License',
    description=(
        'Software for managing music metadata, batch registration '
        'of musical works and royalty processing.'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://dmp.matijakolaric.com',
    project_urls={
        'Created by': 'https://matijakolaric.com',
        'Video Tutorials': 'https://www.youtube.com/watch?v=Tyk6tYuFBrI'
                           '&list=PLQ3e-DuNTFt-mwtKvFLK1euk5uCZdhCUP',
        'Documentation': 'https://django-music-publisher.readthedocs.io/',
        'Code Repository': 'https://github.com/matijakolaric-com/django'
                           '-music-publisher/',
    },
    packages=setuptools.find_packages(exclude=['dmp_project', 'docs']),
    install_requires=(
        'Django>=3.2.12,<4.0',
        'requests>=2.24.0',
    ),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    zip_safe=False,
)



"""Filters used in parsing of CWR files.

"""

from decimal import Decimal, ConversionSyntax, InvalidOperation

from django import template

from music_publisher import models
from music_metadata.territories.territory import Territory

register = template.Library()


@register.filter(name='perc')
def perc(value):
    """Display shares as human-readable string."""
    try:
        value = Decimal(value) / Decimal('100')
        return '{}%'.format(value)
    except (ConversionSyntax, InvalidOperation, TypeError):
        return 'Not convertible to percents.'


@register.filter(name='soc_name')
def soc_name(value):
    """Display society name"""

    value = value.strip().lstrip('0')
    return models.SOCIETY_DICT.get(value, '')


@register.filter(name='capacity')
def capacity(value):
    """Display writer capacity/role"""

    value = value[0:2]
    obj = models.WriterInWork(capacity=value)
    return obj.get_capacity_display()


@register.filter(name='agreement_type')
def agreement_type(value):
    """Display publishing agreement type"""

    value = value.strip()
    return {
        'OG': 'Original general',
        'OS': 'Original specific',
        'PG': 'Sub-Publishing general',
        'PS': 'Sub-Publishing specific',
    }.get(value, 'Unknown')


@register.filter(name='status')
def status(value):
    """Display acknowledgement status"""

    value = value.strip()
    obj = models.WorkAcknowledgement(status=value)
    return obj.get_status_display()


@register.filter(name='flag')
def flag(value):
    """Display flag value"""

    value = value.strip()
    return {
        'Y': 'Yes',
        'N': 'No',
        'U': 'Unknown',
    }.get(value, 'Not set')


@register.filter(name='orimod')
def orimod(value):
    """Display original or modification"""

    value = value.strip()
    return {
        'ORI': 'Original Work',
        'MOD': 'Modification',
    }.get(value, 'Not set')


@register.filter(name='terr')
def terr(value):
    """Display territory"""

    value = value.strip()
    territory = Territory.get(value)
    return str(territory)


@register.filter(name='ie')
def ie(value):
    """Display Included / Excluded"""

    value = value.strip()
    if value == 'E':
        return 'excluding '
    return ''


@register.filter(name='role')
def role(value):
    """Display publisher role/capacity"""

    return {
        'E ': 'Original Publisher',
        'AM': 'Administrator',
        'SE': 'Sub-publisher'
    }.get(value, 'Unknown publisher role')
